# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Constants and type mappings for content models."""

from typing import Type, TypedDict, Union

from typing_extensions import Literal

from .content.audio import AudioMediaContent
from .content.file import FileMediaContent
from .content.image import ImageMediaContent, ImageUrlMediaContent
from .content.text import TextMediaContent
from .content.video import VideoMediaContent

MediaContent = Union[
    TextMediaContent,
    ImageMediaContent,
    ImageUrlMediaContent,
    VideoMediaContent,
    AudioMediaContent,
    FileMediaContent,
]

ContentTypeKey = Literal["text", "image", "image_url", "video", "audio", "file"]
"""Possible content types for the mapping."""


class ContentMappingEntry(TypedDict):
    """TypedDict for content mapping entries.

    Attributes
    ----------
    fields: List[str]
        List of possible field names for the content type
    cls: Type[MediaContent]
        The class to instantiate for this content type
    required_field: str
        The field name required by the class constructor
    """

    fields: list[str]
    cls: Type[MediaContent]
    required_field: str


CONTENT_MAPPING: dict[ContentTypeKey, ContentMappingEntry] = {
    "text": {
        "fields": ["text"],
        "cls": TextMediaContent,
        "required_field": "text",
    },
    "image": {
        "fields": ["image"],
        "cls": ImageMediaContent,
        "required_field": "image",
    },
    "image_url": {
        "fields": ["image_url", "url"],
        "cls": ImageUrlMediaContent,
        "required_field": "image_url",
    },
    "video": {
        "fields": ["video"],
        "cls": VideoMediaContent,
        "required_field": "video",
    },
    "audio": {
        "fields": ["audio"],
        "cls": AudioMediaContent,
        "required_field": "audio",
    },
    "file": {
        "fields": ["file"],
        "cls": FileMediaContent,
        "required_field": "file",
    },
}



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get the standard imports for the flow exporter."""

from typing import Optional

from waldiez.exporting.core import ImportPosition

BUILTIN_IMPORTS = [
    "import csv",
    "import importlib",
    "import json",
    "import os",
    "import sqlite3",
    "import sys",
    "from dataclasses import asdict",
    "from pprint import pprint",
    "from types import ModuleType",
]
TYPING_IMPORT_NAMES = [
    "Annotated",
    "Any",
    "Callable",
    "Dict",
    "List",
    "Optional",
    "Set",
    "Tuple",
    "Union",
]
# pylint: disable=inconsistent-quotes
TYPING_IMPORTS = [f"from typing import {', '.join(TYPING_IMPORT_NAMES)}"]
COMMON_AUTOGEN_IMPORTS = [
    "from autogen import Agent",
    "from autogen import Cache",
    "from autogen import ConversableAgent",
    "from autogen import ChatResult",
    "from autogen import GroupChat",
    "from autogen import runtime_logging",
]


def get_sorted_imports(collected_imports: list[str]) -> list[str]:
    """Get the sorted imports.

    Parameters
    ----------
    collected_imports : list[str]
        The collected imports.

    Returns
    -------
    list[str]
        The sorted imports.
    """
    # Remove duplicates while preserving order by converting to dict
    unique_imports = list(dict.fromkeys(collected_imports))

    sorted_imports = sorted(
        [imp for imp in unique_imports if imp.startswith("import ")]
    ) + sorted([imp for imp in unique_imports if imp.startswith("from ")])
    return sorted_imports


def sort_imports(
    all_imports: list[tuple[str, ImportPosition]],
) -> tuple[list[str], list[str], list[str], list[str], bool]:
    """Sort the imports.

    Parameters
    ----------
    all_imports : list[tuple[str, ImportPosition]]
        All the imports.

    Returns
    -------
    tuple[list[str], list[str], list[str], list[str], bool]
        The sorted imports and a flag if we got `import autogen`.
    """
    builtin_imports: list[str] = BUILTIN_IMPORTS.copy() + TYPING_IMPORTS.copy()
    third_party_imports: list[str] = []
    local_imports: list[str] = []
    autogen_imports: list[str] = COMMON_AUTOGEN_IMPORTS.copy()
    got_import_autogen = False

    for import_string, position in all_imports:
        if "import autogen" in import_string:
            got_import_autogen = True
            continue
        if import_string.startswith("from autogen"):
            autogen_imports.append(import_string)
            continue
        if position == ImportPosition.BUILTINS:
            builtin_imports.append(import_string)
        elif position == ImportPosition.THIRD_PARTY:
            third_party_imports.append(import_string)
        elif position == ImportPosition.LOCAL:  # pragma: no branch
            local_imports.append(import_string)

    autogen_imports = clean_and_group_autogen_imports(autogen_imports)
    third_party_imports = ensure_np_import(third_party_imports)
    sorted_builtins = get_sorted_imports(builtin_imports)
    sorted_third_party = get_sorted_imports(third_party_imports)
    sorted_locals = get_sorted_imports(local_imports)

    return (
        sorted_builtins,
        sorted(autogen_imports),
        sorted_third_party,
        sorted_locals,
        got_import_autogen,
    )


def clean_and_group_autogen_imports(autogen_imports: list[str]) -> list[str]:
    """Cleanup and group autogen imports.

    Parameters
    ----------
    autogen_imports : list[str]
        List of autogen import statements

    Returns
    -------
    list[str]
        Cleaned and grouped autogen imports
    """
    # Group imports by module path
    import_groups: dict[str, set[str]] = {}
    direct_imports: set[str] = set()

    for imp in autogen_imports:
        imp = imp.strip()
        if not imp:
            continue

        if imp.startswith("import autogen"):
            direct_imports.add(imp)
            continue

        # Parse "from autogen.module import items"
        if imp.startswith("from autogen"):  # pragma: no branch
            parts = imp.split(" import ")
            if len(parts) == 2:  # pragma: no branch
                module_path = parts[0]  # "from autogen.module"
                items = parts[1].strip()

                if module_path not in import_groups:
                    import_groups[module_path] = set()

                # Handle multiple imports in one line
                for item in items.split(","):
                    import_groups[module_path].add(item.strip())

    # Build cleaned import list
    cleaned_imports: list[str] = []

    # Add direct imports first
    cleaned_imports.extend(sorted(direct_imports))

    # Add grouped imports, sorted by module path
    for module_path in sorted(import_groups.keys()):
        sorted_items = sorted(import_groups[module_path])
        items_str = ", ".join(sorted_items)
        import_statement = f"{module_path} import {items_str}"
        cleaned_imports.append(import_statement)

    return cleaned_imports


def get_the_imports_string(
    all_imports: list[tuple[str, ImportPosition]],
    is_async: bool,
) -> str:
    """Get the final imports string.

    Parameters
    ----------
    all_imports : list[tuple[str, ImportPosition]]
        All the imports.
    is_async : bool
        If the flow is async.

    Returns
    -------
    str
        The final imports string.
    """
    (
        builtin_imports,
        autogen_imports,
        third_party_imports,
        local_imports,
        got_import_autogen,
    ) = sort_imports(all_imports)

    # Get the final imports string.
    # Making sure that there are two lines
    # after each import section
    # (builtin, third party, local)
    final_string = "\n".join(builtin_imports) + "\n"
    while not final_string.endswith("\n\n"):
        final_string += "\n"

    if is_async:
        final_string += (
            "\nimport aiofiles"
            "\nimport aiosqlite"
            "\nimport anyio"
            "\nimport nest_asyncio"
            "\nfrom aiocsv import AsyncDictWriter\n"
        )

    if got_import_autogen:
        final_string += "\nimport autogen  # type: ignore\n"

    if autogen_imports:  # pragma: no branch
        final_string += "\n".join(autogen_imports) + "\n"

    if third_party_imports:  # pragma: no branch
        final_string += "\n".join(third_party_imports) + "\n"

    while not final_string.endswith("\n\n"):
        final_string += "\n"

    if local_imports:
        final_string += "\n".join(local_imports) + "\n"

    while not final_string.endswith("\n\n"):
        final_string += "\n"

    if is_async:
        final_string += (
            "# pylint: disable=broad-exception-caught\n"
            "try:\n"
            "    nest_asyncio.apply()  # pyright: ignore\n"
            "except BaseException:\n"
            "    pass  # maybe on uvloop?\n"
        )

    return final_string.replace("\n\n\n", "\n\n")  # avoid too many newlines


def ensure_np_import(third_party_imports: list[str]) -> list[str]:
    """Ensure numpy is imported.

    Parameters
    ----------
    third_party_imports : list[str]
        The third party imports.

    Returns
    -------
    list[str]
        The third party imports with numpy.
    """
    if (
        not third_party_imports
        or "import numpy as np" not in third_party_imports
    ):
        third_party_imports.append("import numpy as np")
    return third_party_imports


def gather_imports(
    model_imports: Optional[list[tuple[str, ImportPosition]]] = None,
    tool_imports: Optional[list[tuple[str, ImportPosition]]] = None,
    chat_imports: Optional[list[tuple[str, ImportPosition]]] = None,
    agent_imports: Optional[list[tuple[str, ImportPosition]]] = None,
) -> list[tuple[str, ImportPosition]]:
    """Gather all the imports.

    Parameters
    ----------
    model_imports : tuple[str, ImportPosition]
        The model imports.
    tool_imports : tuple[str, ImportPosition]
        The tool imports.
    chat_imports : tuple[str, ImportPosition]
        The chat imports.
    agent_imports : tuple[str, ImportPosition]
        The agent imports.

    Returns
    -------
    tuple[str, ImportPosition]
        The gathered imports.
    """
    all_imports: list[tuple[str, ImportPosition]] = []
    for import_statement in BUILTIN_IMPORTS:
        all_imports.append(
            (
                import_statement,
                ImportPosition.BUILTINS,
            )
        )
    if model_imports:
        all_imports.extend(model_imports)
    if tool_imports:
        all_imports.extend(tool_imports)
    if chat_imports:
        all_imports.extend(chat_imports)
    if agent_imports:
        all_imports.extend(agent_imports)
    all_imports = deduplicate_imports(all_imports)
    one_line_typing_imports: list[tuple[str, ImportPosition]] = []
    final_imports: list[tuple[str, ImportPosition]] = []
    for import_tuple in all_imports:
        if import_tuple[0].startswith("from typing import "):
            one_line_typing_imports.append(import_tuple)
        else:
            final_imports.append(import_tuple)
    final_typing_imports = get_the_typing_imports(one_line_typing_imports)
    final_imports.insert(1, (final_typing_imports, ImportPosition.BUILTINS))
    return final_imports


def get_the_typing_imports(
    one_line_typing_imports: list[tuple[str, ImportPosition]],
) -> str:
    """Get the typing imports as a single line.

    Parameters
    ----------
    one_line_typing_imports : list[tuple[str, ImportPosition]]
        The one line typing imports.

    Returns
    -------
    str
        The final typing imports string.
    """
    # merge the ones found in the imports
    # with the default ones
    typing_imports: set[str] = set(TYPING_IMPORT_NAMES)
    for import_tuple in one_line_typing_imports:
        if import_tuple[0].startswith(
            "from typing import "
        ):  # pragma: no branch
            # extract the names from the import statement
            names = import_tuple[0].split("from typing import ")[1]
            for name in names.split(","):
                typing_imports.add(name.strip())
    # create the final import statement
    final_typing_imports = (
        f"from typing import {', '.join(sorted(typing_imports))}"
    )
    return final_typing_imports.strip()


def deduplicate_imports(
    imports: list[tuple[str, ImportPosition]],
) -> list[tuple[str, ImportPosition]]:
    """Deduplicate imports while preserving order.

    Parameters
    ----------
    imports : list[tuple[str, ImportPosition]]
        The imports to deduplicate.

    Returns
    -------
    list[tuple[str, ImportPosition]]
        The deduplicated imports.
    """
    # Remove duplicates while preserving order
    seen: set[tuple[str, ImportPosition]] = set()
    deduplicated_imports: list[tuple[str, ImportPosition]] = []
    for import_tuple in imports:
        if import_tuple not in seen:
            seen.add(import_tuple)
            deduplicated_imports.append(import_tuple)
    return deduplicated_imports



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# flake8: noqa: E501
"""Waldiez Model Data."""

from typing import Optional

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..common import WaldiezBase

WaldiezModelAPIType = Literal[
    "openai",
    "azure",
    "deepseek",
    "google",
    "anthropic",
    "mistral",
    "groq",
    "together",
    "nim",
    "cohere",
    "bedrock",
    "other",
]
"""Possible API types for the model."""


class WaldiezModelAWS(WaldiezBase):
    """AWS related parameters.

    Attributes
    ----------
    region : Optional[str]
        The AWS region, by default None.
    access_key : Optional[str]
        The AWS access key, by default None.
    secret_key : Optional[str]
        The AWS secret access key, by default None.
    session_token : Optional[str]
        The AWS session token, by default None.
    profile_name : Optional[str]
        The AWS profile name, by default Nonde.
    """

    region: Annotated[
        Optional[str],
        Field(
            None,
            alias="region",
            title="Region",
            description="The AWS region",
        ),
    ] = None
    access_key: Annotated[
        Optional[str],
        Field(
            None,
            alias="accessKey",
            title="Access Ke",
            description="The AWS access key",
        ),
    ] = None
    secret_key: Annotated[
        Optional[str],
        Field(
            None,
            alias="secretKey",
            title="Secret Key",
            description="The AWS secret key",
        ),
    ] = None
    session_token: Annotated[
        Optional[str],
        Field(
            None,
            alias="sessionToken",
            title="Session Token",
            description="The AWS session token",
        ),
    ] = None
    profile_name: Annotated[
        Optional[str],
        Field(
            None,
            alias="profileName",
            title="Profile Name",
            description="The AWS Profile name to use",
        ),
    ] = None


class WaldiezModelPrice(WaldiezBase):
    """Model Price.

    Attributes
    ----------
    prompt_price_per_1k : float
        The prompt price per 1k tokens.
    completion_token_price_per_1k : float
        The completion token price per 1k tokens.
    """

    prompt_price_per_1k: Annotated[
        Optional[float], Field(None, alias="promptPricePer1k")
    ]
    completion_token_price_per_1k: Annotated[
        Optional[float], Field(None, alias="completionTokenPricePer1k")
    ]


# pylint: disable=line-too-long
class WaldiezModelData(WaldiezBase):
    """Waldiez Model Data.

    Attributes
    ----------
    base_url : Optional[str]
        The base url of the model, by default None.
    api_key : Optional[str]
        The api key to use with the model, by default None.
    api_type : WaldiezModelAPIType
        The api type of the model.
    api_version : Optional[str]
        The api version of the model, by default None.
    temperature : Optional[float]
        The temperature of the model, by default None.
    top_p : Optional[float]
        The top p of the model, by default None.
    max_tokens : Optional[int]
        The max tokens of the model, by default None.
    aws : Optional[WaldiezModelAWS]
    extras: dict[str, str]
        Any extra attributes to include in the LLM Config.
    default_headers : dict[str, str]
        The default headers of the model.
    price : Optional[WaldiezModelPrice]
        The price of the model, by default None.
    """

    base_url: Annotated[
        Optional[str],
        Field(
            default=None,
            title="Base URL",
            description="The base url of the model",
            alias="baseUrl",
        ),
    ] = None
    api_key: Annotated[
        Optional[str],
        Field(
            default=None,
            alias="apiKey",
            title="API Key",
            description="The api key to use with the model",
        ),
    ] = None
    api_type: Annotated[
        WaldiezModelAPIType,
        Field(
            default="other",
            alias="apiType",
            title="API Type",
            description="The api type of the model",
        ),
    ] = "other"
    api_version: Annotated[
        Optional[str],
        Field(
            default=None,
            alias="apiVersion",
            title="API Version",
            description="The api version of the model",
        ),
    ] = None
    temperature: Annotated[
        Optional[float],
        Field(
            default=None,
            alias="temperature",
            title="Temperature",
            description="The temperature of the model",
        ),
    ] = None
    top_p: Annotated[
        Optional[float],
        Field(
            default=None,
            alias="topP",
            title="Top P",
            description="The top p of the model",
        ),
    ] = None
    max_tokens: Annotated[
        Optional[int],
        Field(
            default=None,
            alias="maxTokens",
            title="Max Tokens",
            description="The max tokens of the model",
        ),
    ] = None
    aws: Annotated[
        Optional[WaldiezModelAWS],
        Field(
            default=None,
            alias="aws",
            title="AWS",
            description="The AWS related parameters",
        ),
    ] = None
    extras: Annotated[
        dict[str, str],
        Field(
            alias="extras",
            default_factory=dict,
            title="Extras",
            description="Any extra attributes to include in the LLM Config",
        ),
    ] = {}
    default_headers: Annotated[
        dict[str, str],
        Field(
            alias="defaultHeaders",
            default_factory=dict,
            title="Default Headers",
            description="The default headers of the model",
        ),
    ] = {}
    price: Annotated[
        Optional[WaldiezModelPrice],
        Field(
            default=None, title="Price", description="The price of the model"
        ),
    ] = None



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Get comment string for scripts or notebooks."""


def get_comment(
    comment: str,
    for_notebook: bool = False,
    md_headings: int = 3,
) -> str:
    """Get a comment for the script or notebook.

    Parameters
    ----------
    comment : str
        The comment to add.
    for_notebook : bool, optional
        Whether the comment is for a notebook, by default False
    md_headings : int, optional
        The number of markdown headings to use (if for_notebook is True),
        by default 2

    Returns
    -------
    str
        The formatted comment.
    """
    if for_notebook:
        # For notebooks, we use markdown headings
        heading = "#" + "#" * md_headings
        return f"# %% [markdown]\n{heading} {comment}\n# %% \n"
    # For scripts, we use a simple comment
    return f"# {comment}\n"



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Run a waldiez flow.

The flow is first converted to an autogen flow with agents, chats, and tools.
We then chown to temporary directory, call the flow's `main()` and
return the results. Before running the flow, any additional environment
variables specified in the waldiez file are set.
"""

# pylint: disable=import-outside-toplevel,reimported

import gc
import importlib.util
import sys
import tempfile
from pathlib import Path
from types import ModuleType, TracebackType
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    Set,
    Type,
    Union,
)

from .exporter import WaldiezExporter
from .io import StructuredIOStream
from .models.waldiez import Waldiez
from .running import (
    a_chdir,
    a_install_requirements,
    after_run,
    before_run,
    chdir,
    get_printer,
    install_requirements,
    refresh_environment,
    reset_env_vars,
    set_env_vars,
)

if TYPE_CHECKING:
    from autogen import ChatResult  # type: ignore


class WaldiezRunner:
    """Waldiez runner class."""

    def __init__(
        self, waldiez: Waldiez, file_path: Optional[Union[str, Path]] = None
    ) -> None:
        """Initialize the Waldiez manager."""
        self._waldiez = waldiez
        self._running = False
        self._file_path = file_path
        self._exporter = WaldiezExporter(waldiez)
        self._called_install_requirements = False

    @classmethod
    def load(
        cls,
        waldiez_file: Union[str, Path],
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[list[str]] = None,
        requirements: Optional[list[str]] = None,
    ) -> "WaldiezRunner":
        """Create a WaldiezRunner instance from a file.

        Parameters
        ----------
        waldiez_file : Union[str, Path]
            The file path.
        name : Optional[str], optional
            The name of the Waldiez, by default None.
        description : Optional[str], optional
            The description of the Waldiez, by default None.
        tags : Optional[list[str]], optional
            The tags of the Waldiez, by default None.
        requirements : Optional[list[str]], optional
            The requirements of the Waldiez, by default None.

        Returns
        -------
        WaldiezRunner
            The Waldiez runner instance.

        Raises
        ------
        FileNotFoundError
            If the file is not found.
        RuntimeError
            If the file is not a valid Waldiez file.
        """
        waldiez = Waldiez.load(
            waldiez_file,
            name=name,
            description=description,
            tags=tags,
            requirements=requirements,
        )
        return cls(waldiez, file_path=waldiez_file)

    def __enter__(
        self,
    ) -> "WaldiezRunner":
        """Enter the context manager."""
        return self

    async def __aenter__(
        self,
    ) -> "WaldiezRunner":
        """Enter the context manager asynchronously."""
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        """Exit the context manager."""
        if self._running:
            self._running = False

    async def __aexit__(
        self,
        exc_type: Type[BaseException],
        exc_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        """Exit the context manager asynchronously."""
        if self._running:
            self._running = False

    @property
    def waldiez(self) -> Waldiez:
        """Get the Waldiez instance."""
        return self._waldiez

    @property
    def is_async(self) -> bool:
        """Check if the workflow is async."""
        return self.waldiez.is_async

    @property
    def running(self) -> bool:
        """Get the running status."""
        return self._running

    def gather_requirements(self) -> Set[str]:
        """Gather extra requirements to install before running the flow.

        Returns
        -------
        Set[str]
            The extra requirements.
        """
        extra_requirements = {
            req for req in self.waldiez.requirements if req not in sys.modules
        }
        return extra_requirements

    def install_requirements(self) -> None:
        """Install the requirements for the flow."""
        self._called_install_requirements = True
        printer = get_printer()
        extra_requirements = self.gather_requirements()
        if extra_requirements:
            install_requirements(extra_requirements, printer)

    async def a_install_requirements(self) -> None:
        """Install the requirements for the flow asynchronously."""
        self._called_install_requirements = True
        printer = get_printer()
        extra_requirements = self.gather_requirements()
        if extra_requirements:
            await a_install_requirements(extra_requirements, printer)

    def _before_run(
        self,
        temp_dir: Path,
        file_name: str,
        module_name: str,
        printer: Callable[..., None],
    ) -> tuple[ModuleType, dict[str, str]]:
        self._exporter.export(Path(file_name))
        # unique_names = self._exporter.context.get_unique_names()
        spec = importlib.util.spec_from_file_location(
            module_name, temp_dir / file_name
        )
        if not spec or not spec.loader:
            raise ImportError("Could not import the flow")
        sys.path.insert(0, str(temp_dir))
        old_vars = set_env_vars(self.waldiez.get_flow_env_vars())
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        printer("<Waldiez> - Starting workflow...")
        printer(self.waldiez.info.model_dump_json())
        return module, old_vars

    def _run(
        self,
        output_path: Optional[Union[str, Path]],
        uploads_root: Optional[Union[str, Path]],
        use_structured_io: bool = False,
        skip_mmd: bool = False,
    ) -> Union["ChatResult", list["ChatResult"], dict[int, "ChatResult"]]:
        """Run the Waldiez workflow.

        Parameters
        ----------
        output_path : Optional[Union[str, Path]]
            The output path.
        uploads_root : Optional[Union[str, Path]]
            The runtime uploads root.
        use_structured_io : bool
            Whether to use structured IO instead of the default 'input/print'.
        skip_mmd : bool
            Whether to skip the Mermaid diagram generation.

        Returns
        -------
        Union[ChatResult, list[ChatResult]]
            The result(s) of the chat(s).
        """
        temp_dir = Path(tempfile.mkdtemp())
        file_name = before_run(output_path, uploads_root)
        module_name = file_name.replace(".py", "")
        if not self._called_install_requirements:
            self.install_requirements()
        refresh_environment()
        printer = get_printer()
        printer(
            "Requirements installed.\n"
            "NOTE: If new packages were added and you are using Jupyter, "
            "you might need to restart the kernel."
        )
        results: Union[
            "ChatResult", list["ChatResult"], dict[int, "ChatResult"]
        ] = []
        with chdir(to=temp_dir):
            module, old_vars = self._before_run(
                temp_dir=temp_dir,
                file_name=file_name,
                module_name=module_name,
                printer=printer,
            )
            if use_structured_io:
                stream = StructuredIOStream(
                    timeout=120,  # 2 minutes
                    uploads_root=uploads_root,
                )
                with StructuredIOStream.set_default(stream):
                    results = module.main()
            else:
                results = module.main()
            printer("<Waldiez> - Workflow finished")
            sys.path.pop(0)
            reset_env_vars(old_vars)
        after_run(
            temp_dir=temp_dir,
            output_path=output_path,
            printer=printer,
            flow_name=self.waldiez.name,
            skip_mmd=skip_mmd,
        )
        gc.collect()
        refresh_environment()
        return results

    async def _a_run(
        self,
        output_path: Optional[Union[str, Path]],
        uploads_root: Optional[Union[str, Path]],
        use_structured_io: bool = False,
        skip_mmd: bool = False,
    ) -> Union["ChatResult", list["ChatResult"], dict[int, "ChatResult"]]:
        """Run the Waldiez workflow asynchronously."""
        temp_dir = Path(tempfile.mkdtemp())
        file_name = before_run(output_path, uploads_root)
        module_name = file_name.replace(".py", "")
        if not self._called_install_requirements:
            await self.a_install_requirements()
        refresh_environment()
        printer = get_printer()
        printer(
            "Requirements installed.\n"
            "NOTE: If new packages were added and you are using Jupyter, "
            "you might need to restart the kernel."
        )
        results: Union[
            "ChatResult", list["ChatResult"], dict[int, "ChatResult"]
        ] = []
        async with a_chdir(to=temp_dir):
            module, old_vars = self._before_run(
                temp_dir=temp_dir,
                file_name=file_name,
                module_name=module_name,
                printer=printer,
            )
            if use_structured_io:
                stream = StructuredIOStream(
                    timeout=120,  # 2 minutes
                    uploads_root=uploads_root,
                )
                with StructuredIOStream.set_default(stream):
                    results = await module.main()
            else:
                results = await module.main()
            sys.path.pop(0)
            reset_env_vars(old_vars)
        after_run(
            temp_dir=temp_dir,
            output_path=output_path,
            printer=printer,
            flow_name=self.waldiez.name,
            skip_mmd=skip_mmd,
        )
        gc.collect()
        refresh_environment()
        return results

    def run(
        self,
        output_path: Optional[Union[str, Path]] = None,
        uploads_root: Optional[Union[str, Path]] = None,
        use_structured_io: bool = False,
        skip_mmd: bool = False,
    ) -> Union["ChatResult", list["ChatResult"], dict[int, "ChatResult"]]:
        """Run the Waldiez workflow.

        Parameters
        ----------
        output_path : Optional[Union[str, Path]], optional
            The output path, by default None.
        uploads_root : Optional[Union[str, Path]], optional
            The uploads root, to get user-uploaded files, by default None.
        use_structured_io : bool, optional
            Whether to use structured IO instead of the default 'input/print',
            by default False.
        skip_mmd : bool, optional
            Whether to skip the Mermaid diagram generation, by default False.

        Returns
        -------
        Union["ChatResult", list["ChatResult"], dict[int, "ChatResult"]]
            The result(s) of the chat(s).

        Raises
        ------
        RuntimeError
            If the workflow is already running.
        """
        if self.waldiez.is_async:
            # pylint: disable=import-outside-toplevel
            from anyio.from_thread import start_blocking_portal

            with start_blocking_portal(backend="asyncio") as portal:
                return portal.call(
                    self._a_run,
                    output_path,
                    uploads_root,
                    use_structured_io,
                    skip_mmd,
                )
        if self._running is True:
            raise RuntimeError("Workflow already running")
        self._running = True
        file_path = output_path or self._file_path
        try:
            return self._run(
                file_path,
                uploads_root=uploads_root,
                use_structured_io=use_structured_io,
                skip_mmd=skip_mmd,
            )
        finally:
            self._running = False

    async def a_run(
        self,
        output_path: Optional[Union[str, Path]] = None,
        uploads_root: Optional[Union[str, Path]] = None,
        use_structured_io: bool = False,
        skip_mmd: bool = False,
    ) -> Union["ChatResult", list["ChatResult"], dict[int, "ChatResult"]]:
        """Run the Waldiez workflow asynchronously.

        Parameters
        ----------
        output_path : Optional[Union[str, Path]], optional
            The output path, by default None.
        uploads_root : Optional[Union[str, Path]], optional
            The uploads root, to get user-uploaded files, by default None.
        use_structured_io : bool, optional
            Whether to use structured IO instead of the default 'input/print',
            by default False.
        skip_mmd : bool, optional
            Whether to skip the Mermaid diagram generation, by default False.

        Returns
        -------
        Union[ChatResult, list[ChatResult]], dict[int, ChatResult]
            The result(s) of the chat(s).

        Raises
        ------
        RuntimeError
            If the workflow is already running.
        """
        if self._running is True:
            raise RuntimeError("Workflow already running")
        self._running = True
        file_path = output_path or self._file_path
        try:
            return await self._a_run(
                file_path,
                uploads_root=uploads_root,
                use_structured_io=use_structured_io,
                skip_mmd=skip_mmd,
            )
        finally:
            self._running = False



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Utility functions for Waldiez exporting."""

from .comment import get_comment
from .llm_config import get_agent_llm_config_arg

__all__ = [
    "get_comment",
    "get_agent_llm_config_arg",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods
"""Chat specific extras module."""

from dataclasses import dataclass

from ..enums import AgentPosition, ContentOrder, ExportPosition
from ..result import ExportResult
from .base import BaseExtras


@dataclass
class ChatExtras(BaseExtras):
    """Extras for chat exporters.

    Attributes
    ----------
    chat_definition : str
        The chat definition content.
    chat_initiation : str
        The chat initiation content.
    """

    chat_prerequisites: str = ""
    chat_initiation: str = ""
    chat_registration: str = ""

    def set_chat_prerequisites(self, prerequisites: str) -> None:
        """Set the chat prerequisites.

        Parameters
        ----------
        prerequisites : str
            The chat prerequisites content.
        """
        self.chat_prerequisites = prerequisites

    def set_chat_initiation(self, initiation: str) -> None:
        """Set the chat initiation.

        Parameters
        ----------
        initiation : str
            The chat initiation content.
        """
        self.chat_initiation = initiation

    def set_chat_registration(self, registration: str) -> None:
        """Set the chat registration.

        Parameters
        ----------
        registration : str
            The chat registration content.
        """
        self.chat_registration = registration

    def add_registration(self, registration: str) -> None:
        """Add chat registration content.

        Parameters
        ----------
        registration : str
            The chat registration content.
        """
        if registration and registration.strip():
            self.chat_registration += "\n" + registration.strip() + "\n"

    def has_specific_content(self) -> bool:
        """Check for chat specific content.

        Returns
        -------
        bool
            True if there's chat specific content.
        """
        return bool(
            self.chat_initiation.strip()
            or self.chat_prerequisites.strip()
            or self.chat_registration.strip()
        )

    def _contribute_specific_content(self, result: ExportResult) -> None:
        """Contribute chat specific content.

        Parameters
        ----------
        result : ExportResult
            The export result to contribute to.
        """
        # Chat prerequisites go before the chat definition
        if self.chat_prerequisites:
            result.add_content(
                self.chat_prerequisites,
                ExportPosition.AGENTS,
                order=ContentOrder.LATE_CLEANUP,
                skip_strip=True,
            )

        # Chat initiation goes inside "def main()"
        if self.chat_initiation:  # pragma: no branch
            result.add_content(
                self.chat_initiation,
                ExportPosition.CHATS,
                order=ContentOrder.MAIN_CONTENT,
                skip_strip=True,
            )

        # Chat registration goes after all agents are defined
        # (agent.register_nested_chat..))
        if self.chat_registration:  # pragma: no branch
            result.add_content(
                self.chat_registration,
                ExportPosition.AGENTS,
                agent_position=AgentPosition.AFTER_ALL,
                order=ContentOrder.POST_CONTENT,
            )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""RAG user agent retrieve config."""

import os
from pathlib import Path
from typing import Optional, Union

from pydantic import Field, model_validator
from typing_extensions import Annotated, Literal, Self

from ...common import WaldiezBase, check_function, generate_function
from .vector_db_config import WaldiezRagUserProxyVectorDbConfig

WaldiezRagUserProxyTask = Literal["code", "qa", "default"]
"""Possible tasks for the retrieve chat."""
WaldiezRagUserProxyVectorDb = Literal["chroma", "pgvector", "mongodb", "qdrant"]
"""Possible vector dbs for the retrieve chat."""
WaldiezRagUserProxyChunkMode = Literal["multi_lines", "one_line"]
"""Possible chunk modes for the retrieve chat."""
WaldiezRagUserProxyModels: dict[WaldiezRagUserProxyVectorDb, str] = {
    "chroma": "all-MiniLM-L6-v2",
    "mongodb": "all-MiniLM-L6-v2",
    "pgvector": "all-MiniLM-L6-v2",
    "qdrant": "BAAI/bge-small-en-v1.5",
}

CUSTOM_EMBEDDING_FUNCTION = "custom_embedding_function"
CUSTOM_EMBEDDING_FUNCTION_ARGS: list[str] = []
CUSTOM_EMBEDDING_FUNCTION_TYPES: tuple[list[str], str] = (
    [],
    "Callable[..., Any]",
)
CUSTOM_TOKEN_COUNT_FUNCTION = "custom_token_count_function"  # nosec
CUSTOM_TOKEN_COUNT_FUNCTION_ARGS = ["text", "model"]  # nosec
CUSTOM_TOKEN_COUNT_FUNCTION_TYPES = (
    ["str", "str"],
    "int",
)

CUSTOM_TEXT_SPLIT_FUNCTION = "custom_text_split_function"
CUSTOM_TEXT_SPLIT_FUNCTION_ARGS = [
    "text",
    "max_tokens",
    "chunk_mode",
    "must_break_at_empty_line",
    "overlap",
]
CUSTOM_TEXT_SPLIT_FUNCTION_TYPES = (
    ["str", "int", "str", "bool", "int"],
    "list[str]",
)
NOT_LOCAL = (
    "http://",
    "https://",
    "ftp://",
    "ftps://",
    "sftp://",
    "hdfs",
    "s3://",
    "gs://",
    "azure://",
)


class WaldiezRagUserProxyRetrieveConfig(WaldiezBase):
    """RAG user agent.

    Attributes
    ----------
    task : Literal["code", "qa", "default"]
        The task of the retrieve chat.
        Possible values are 'code', 'qa' and 'default'.
        System prompt will be different for different tasks.
        The default value is default, which supports both code and qa,
        and provides source information in the end of the response.
    vector_db : Literal["chroma", "pgvector", "mongodb", "qdrant"]
        The vector db for the retrieve chat.
    db_config : Annotated[WaldiezVectorDbConfig, Field]
        The config for the selected vector db.
    docs_path : Optional[Union[str, list[str]]]
        The path to the docs directory. It can also be the path to a single
        file, the url to a single file or a list of directories, files and
        urls. Default is None, which works only if the collection is already
        created.
    new_docs : bool
        When True, only adds new documents to the collection; when False,
        updates existing documents and adds new ones. Default is True.
        Document id is used to determine if a document is new or existing.
        By default, the id is the hash value of the content.
    model : Optional[str]
        The model to use for the retrieve chat. If key not provided, a default
        model gpt-4 will be used.
    chunk_token_size : Optional[int]
        The chunk token size for the retrieve chat. If key not provided, a
        default size max_tokens * 0.4 will be used.
    context_max_tokens : Optional[int]
        The context max token size for the retrieve chat. If key not provided,
        a default size max_tokens * 0.8 will be used.
    chunk_mode : Optional[str]
        The chunk mode for the retrieve chat. Possible values are 'multi_lines'
        and 'one_line'. If key not provided, a default mode multi_lines will be
        used.
    must_break_at_empty_line : bool
        Chunk will only break at empty line if True. Default is True. If
        chunk_mode is 'one_line', this parameter will be ignored.
    use_custom_embedding: bool
        Whether to use custom embedding for the retrieve chat. Default is False.
        If True, the embedding_function should be provided.
    embedding_function : Optional[str]
        The embedding function for creating the vector db. Default is None,
        SentenceTransformer with the given embedding_model will be used. If
        you want to use OpenAI, Cohere, HuggingFace or other embedding
        functions, you can pass it here, follow the examples in
        https://docs.trychroma.com/guides/embeddings.
    customized_prompt : Optional[str]
        The customized prompt for the retrieve chat. Default is None.
    customized_answer_prefix : Optional[str]
        The customized answer prefix for the retrieve chat. Default is ''. If
        not '' and the customized_answer_prefix is not in the answer, Update
        Context will be triggered.
    update_context : bool
        If False, will not apply Update Context for interactive retrieval.
        Default is True.
    collection_name : Optional[str]
        The name of the collection. If key not provided, a default name
        autogen-docs will be used.
    get_or_create : bool
        Whether to get the collection if it exists. Default is False.
    overwrite : bool
        Whether to overwrite the collection if it exists. Default is False.
        Case 1. if the collection does not exist, create the collection. Case
        2. the collection exists, if overwrite is True, it will overwrite the
        collection. Case 3. the collection exists and overwrite is False, if
        get_or_create is True, it will get the collection, otherwise it raise a
        ValueError.
    use_custom_token_count: bool
        Whether to use custom token count function for the retrieve chat.
        Default is False. If True, the custom_token_count_function should be
        provided.
    custom_token_count_function : Optional[str]
        A custom function to count the number of tokens in a string. The
        function should take (text:str, model:str) as input and return the
        token_count(int). the retrieve_config['model'] will be passed in the
        function. Default is autogen.token_count_utils.count_token that uses
        tiktoken, which may not be accurate for non-OpenAI models.
    use_custom_text_split: bool
        Whether to use custom text split function for the retrieve chat. Default
        is False. If True, the custom_text_split_function should be provided.
    custom_text_split_function : Optional[str]
        A custom function to split a string into a list of strings. Default is
        None, will use the default function in autogen.retrieve_utils.
        split_text_to_chunks.
    custom_text_types : Optional[list[str]]
        A list of file types to be processed. Default is autogen.retrieve_utils.
        TEXT_FORMATS. This only applies to files under the directories in
        docs_path. Explicitly included files and urls will be chunked
        regardless of their types.
    recursive : bool
        Whether to search documents recursively in the docs_path. Default is
        True.
    distance_threshold : float
        The threshold for the distance score, only distance smaller than it
        will be returned. Will be ignored if < 0. Default is -1.
    embedding_function_string : Optional[str]
        The embedding function string (if use_custom_embedding is True).
    token_count_function_string : Optional[str]
        The token count function string (if use_custom_token_count is True).
    text_split_function_string : Optional[str]
        The text split function string (if use_custom_text_split is True).
    n_results: Optional[int]
        The number of results to return. Default is None, which will return all

    Functions
    ---------
    validate_custom_embedding_function
        Validate the custom embedding function.
    validate_custom_token_count_function
        Validate the custom token count function.
    validate_custom_text_split_function
        Validate the custom text split function.
    validate_rag_user_data
        Validate the RAG user data.
    """

    task: Annotated[
        WaldiezRagUserProxyTask,
        Field(
            default="default",
            title="Task",
            description=(
                "The task of the retrieve chat. "
                "Possible values are 'code', 'qa' and 'default'. "
                "System prompt will be different for different tasks. "
                "The default value is default, which supports both code, "
                "and qa and provides source information in the end of "
                "the response."
            ),
        ),
    ] = "default"
    vector_db: Annotated[
        WaldiezRagUserProxyVectorDb,
        Field(
            default="chroma",
            title="Vector DB",
            description="The vector db for the retrieve chat.",
        ),
    ] = "chroma"
    db_config: Annotated[
        WaldiezRagUserProxyVectorDbConfig,
        Field(
            title="DB Config",
            description="The config for the selected vector db.",
            default_factory=WaldiezRagUserProxyVectorDbConfig,
        ),
    ]
    docs_path: Annotated[
        Optional[Union[str, list[str]]],
        Field(
            default=None,
            title="Docs Path",
            description=(
                "The path to the docs directory. It can also be the path to "
                "a single file, the url to a single file or a list of "
                "directories, files and urls. Default is None, which works "
                "only if the collection is already created."
            ),
        ),
    ] = None
    new_docs: Annotated[
        bool,
        Field(
            default=True,
            title="New Docs",
            description=(
                "When True, only adds new documents to the collection; "
                "when False, updates existing documents and adds new ones. "
                "Default is True. Document id is used to determine if a "
                "document is new or existing. By default, the id is the "
                "hash value of the content."
            ),
        ),
    ] = True
    model: Annotated[
        Optional[str],
        Field(
            default=None,
            title="Model",
            description=(
                "The model to use for the retrieve chat. If key not provided, "
                "we check for models linked to the agent."
            ),
        ),
    ] = None
    chunk_token_size: Annotated[
        Optional[int],
        Field(
            default=None,
            title="Chunk Token Size",
            description=(
                "The chunk token size for the retrieve chat.  "
                "If key not provided, a default size max_tokens * 0.4 "
                "will be used."
            ),
        ),
    ] = None
    context_max_tokens: Annotated[
        Optional[int],
        Field(
            default=None,
            title="Context Max Tokens",
            description=(
                "The context max token size for the retrieve chat. "
                "If key not provided, a default size max_tokens * 0.8 "
                "will be used."
            ),
        ),
    ] = None
    chunk_mode: Annotated[
        WaldiezRagUserProxyChunkMode,
        Field(
            default="multi_lines",
            title="Chunk Mode",
            description=(
                "The chunk mode for the retrieve chat. Possible values are "
                "'multi_lines' and 'one_line'. If key not provided, "
                "a default mode multi_lines will be used."
            ),
        ),
    ] = "multi_lines"
    must_break_at_empty_line: Annotated[
        bool,
        Field(
            default=True,
            title="Must Break at Empty Line",
            description=(
                "Chunk will only break at empty line if True. Default is True. "
                "If chunk_mode is 'one_line', this parameter will be ignored."
            ),
        ),
    ] = True
    use_custom_embedding: Annotated[
        bool,
        Field(
            default=False,
            title="Use Custom Embedding",
            description=(
                "Whether to use custom embedding for the retrieve chat. "
                "Default is False. If True, the embedding_function should be "
                "provided."
            ),
        ),
    ] = False
    embedding_function: Annotated[
        Optional[str],
        Field(
            default=None,
            title="Embedding Function",
            description=(
                "The embedding function for creating the vector db. "
                "Default is None, SentenceTransformer with the given "
                "embedding_model will be used. If you want to use OpenAI, "
                "Cohere, HuggingFace or other embedding functions, "
                "you can pass it here, follow the examples in "
                "https://docs.trychroma.com/guides/embeddings."
            ),
        ),
    ] = None
    customized_prompt: Annotated[
        Optional[str],
        Field(
            default=None,
            title="Customized Prompt",
            description=(
                "The customized prompt for the retrieve chat. Default is None."
            ),
        ),
    ] = None
    customized_answer_prefix: Annotated[
        Optional[str],
        Field(
            default="",
            title="Customized Answer Prefix",
            description=(
                "The customized answer prefix for the retrieve chat. "
                "Default is ''. If not '' and the customized_answer_prefix is "
                "not in the answer, Update Context will be triggered."
            ),
        ),
    ] = ""
    update_context: Annotated[
        bool,
        Field(
            default=True,
            title="Update Context",
            description=(
                "If False, will not apply Update Context for interactive "
                "retrieval. Default is True."
            ),
        ),
    ] = True
    collection_name: Annotated[
        str,
        Field(
            default="autogen-docs",
            title="Collection Name",
            description=(
                "The name of the collection. If key not provided, "
                "a default name autogen-docs will be used."
            ),
        ),
    ] = "autogen-docs"
    get_or_create: Annotated[
        bool,
        Field(
            default=False,
            title="Get or Create",
            description=(
                "Whether to get the collection if it exists. Default is False."
            ),
        ),
    ] = False
    overwrite: Annotated[
        bool,
        Field(
            default=False,
            title="Overwrite",
            description=(
                "Whether to overwrite the collection if it exists. "
                "Default is False. "
                "Case 1. if the collection does not exist,"
                " create the collection. "
                "Case 2. the collection exists, if overwrite is True,"
                " it will overwrite the collection. "
                "Case 3. the collection exists and overwrite is False, if"
                " get_or_create is True, it will get the collection,"
                " otherwise it raise a ValueError."
            ),
        ),
    ] = False
    use_custom_token_count: Annotated[
        bool,
        Field(
            default=False,
            title="Use Custom Token Count",
            description=(
                "Whether to use custom token count function for the retrieve "
                "chat. Default is False. If True, the "
                "custom_token_count_function should be provided."
            ),
        ),
    ] = False
    custom_token_count_function: Annotated[
        Optional[str],
        Field(
            default=None,
            title="Custom Token Count Function",
            description=(
                "A custom function to count the number of tokens in a string. "
                "The function should take (text:str, model:str) as input "
                "and return the token_count(int). the retrieve_config['model'] "
                "will be passed in the function. "
                "Default is autogen.token_count_utils.count_token that uses "
                "tiktoken, which may not be accurate for non-OpenAI models."
            ),
        ),
    ] = None
    use_custom_text_split: Annotated[
        bool,
        Field(
            default=False,
            title="Use Custom Text Split",
            description=(
                "Whether to use custom text split function for the retrieve "
                "chat. Default is False. If True, the "
                "custom_text_split_function should be provided."
            ),
        ),
    ] = False
    custom_text_split_function: Annotated[
        Optional[str],
        Field(
            default=None,
            title="Custom Text Split Function",
            description=(
                "A custom function to split a string into a list of strings. "
                "Default is None, will use the default function in "
                "autogen.retrieve_utils.split_text_to_chunks."
            ),
        ),
    ] = None
    custom_text_types: Annotated[
        Optional[list[str]],
        Field(
            default=None,
            title="Custom Text Types",
            description=(
                "A list of file types to be processed. "
                "Default is autogen.retrieve_utils.TEXT_FORMATS. "
                "This only applies to files under the directories in "
                "docs_path. Explicitly included files and urls will be "
                "chunked regardless of their types."
            ),
        ),
    ] = None
    recursive: Annotated[
        bool,
        Field(
            default=True,
            title="Recursive",
            description=(
                "Whether to search documents recursively in the docs_path. "
                "Default is True."
            ),
        ),
    ] = True
    distance_threshold: Annotated[
        float,
        Field(
            default=-1,
            title="Distance Threshold",
            description=(
                "The threshold for the distance score, only distance"
                " smaller than this will be returned. "
                "Will be ignored if < 0. Default is -1."
            ),
        ),
    ] = -1
    n_results: Annotated[
        Optional[int],
        Field(
            default=None,
            title="Number of Results",
            description=(
                "The number of results to return. Default is None, "
                "which will return all."
                "Use None or <1 to return all results."
            ),
        ),
    ] = None
    _embedding_function_string: Optional[str] = None

    _token_count_function_string: Optional[str] = None

    _text_split_function_string: Optional[str] = None

    @property
    def embedding_function_string(self) -> Optional[str]:
        """Get the embedding function string.

        Returns
        -------
        Optional[str]
            The embedding function string.
        """
        return self._embedding_function_string

    @property
    def token_count_function_string(self) -> Optional[str]:
        """Get the token count function string.

        Returns
        -------
        Optional[str]
            The token count function string.
        """
        return self._token_count_function_string

    @property
    def text_split_function_string(self) -> Optional[str]:
        """Get the text split function string.

        Returns
        -------
        Optional[str]
            The text split function string.
        """
        return self._text_split_function_string

    def get_custom_embedding_function(
        self,
        name_prefix: Optional[str] = None,
        name_suffix: Optional[str] = None,
    ) -> tuple[str, str]:
        """Generate the custom embedding function.

        Parameters
        ----------
        name_prefix : str
            The function name prefix.
        name_suffix : str
            The function name suffix.

        Returns
        -------
        tuple[str, str]
            The custom embedding function and the function name.
        """
        function_name = CUSTOM_EMBEDDING_FUNCTION
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        return (
            generate_function(
                function_name=function_name,
                function_args=CUSTOM_EMBEDDING_FUNCTION_ARGS,
                function_types=CUSTOM_EMBEDDING_FUNCTION_TYPES,
                function_body=self.embedding_function_string or "",
            ),
            function_name,
        )

    def get_custom_token_count_function(
        self,
        name_prefix: Optional[str] = None,
        name_suffix: Optional[str] = None,
    ) -> tuple[str, str]:
        """Generate the custom token count function.

        Parameters
        ----------
        name_prefix : str
            The function name prefix.
        name_suffix : str
            The function name suffix.

        Returns
        -------
        tuple[str, str]
            The custom token count function and the function name.
        """
        function_name = CUSTOM_TOKEN_COUNT_FUNCTION
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        return (
            generate_function(
                function_name=function_name,
                function_args=CUSTOM_TOKEN_COUNT_FUNCTION_ARGS,
                function_types=CUSTOM_TOKEN_COUNT_FUNCTION_TYPES,
                function_body=self.token_count_function_string or "",
            ),
            function_name,
        )

    def get_custom_text_split_function(
        self,
        name_prefix: Optional[str] = None,
        name_suffix: Optional[str] = None,
    ) -> tuple[str, str]:
        """Generate the custom text split function.

        Parameters
        ----------
        name_prefix : str
            The function name prefix.
        name_suffix : str
            The function name suffix.

        Returns
        -------
        tuple[str, str]
            The custom text split function and the function name.
        """
        function_name = CUSTOM_TEXT_SPLIT_FUNCTION
        if name_prefix:
            function_name = f"{name_prefix}_{function_name}"
        if name_suffix:
            function_name = f"{function_name}_{name_suffix}"
        return (
            generate_function(
                function_name=function_name,
                function_args=CUSTOM_TEXT_SPLIT_FUNCTION_ARGS,
                function_types=CUSTOM_TEXT_SPLIT_FUNCTION_TYPES,
                function_body=self.text_split_function_string or "",
            ),
            function_name,
        )

    def validate_custom_embedding_function(self) -> None:
        """Validate the custom embedding function.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if self.use_custom_embedding:
            if not self.embedding_function:
                raise ValueError(
                    "The embedding_function is required "
                    "if use_custom_embedding is True."
                )
            valid, error_or_content = check_function(
                code_string=self.embedding_function,
                function_name=CUSTOM_EMBEDDING_FUNCTION,
                function_args=CUSTOM_EMBEDDING_FUNCTION_ARGS,
            )
            if not valid:
                raise ValueError(error_or_content)
            self._embedding_function_string = error_or_content

    def validate_custom_token_count_function(self) -> None:
        """Validate the custom token count function.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if self.use_custom_token_count:
            if not self.custom_token_count_function:
                raise ValueError(
                    "The custom_token_count_function is required "
                    "if use_custom_token_count is True."
                )
            valid, error_or_content = check_function(
                code_string=self.custom_token_count_function,
                function_name=CUSTOM_TOKEN_COUNT_FUNCTION,
                function_args=CUSTOM_TOKEN_COUNT_FUNCTION_ARGS,
            )
            if not valid:
                raise ValueError(error_or_content)
            self._token_count_function_string = error_or_content

    def validate_custom_text_split_function(self) -> None:
        """Validate the custom text split function.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if self.use_custom_text_split:
            if not self.custom_text_split_function:
                raise ValueError(
                    "The custom_text_split_function is required "
                    "if use_custom_text_split is True."
                )
            valid, error_or_content = check_function(
                code_string=self.custom_text_split_function,
                function_name=CUSTOM_TEXT_SPLIT_FUNCTION,
                function_args=CUSTOM_TEXT_SPLIT_FUNCTION_ARGS,
            )
            if not valid:
                raise ValueError(error_or_content)
            self._text_split_function_string = error_or_content

    def validate_docs_path(self) -> None:
        """Validate the docs path.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if not self.docs_path:
            return

        # Normalize to list
        doc_paths = (
            [self.docs_path]
            if isinstance(self.docs_path, str)
            else self.docs_path
        )

        validated_paths: list[str] = []

        for path in doc_paths:
            # Skip duplicates
            if path in validated_paths:
                continue

            # Check if it's a remote path
            is_remote = is_remote_path(path)
            if is_remote:
                # Remote paths: ensure proper raw string wrapping if needed
                content = extract_raw_string_content(path)
                validated_paths.append(f'r"{content}"')
                continue

            # Handle local paths
            # First remove any file:// scheme
            cleaned_path = remove_file_scheme(path)
            content = extract_raw_string_content(cleaned_path)

            # Determine if it's likely a folder
            is_folder = string_represents_folder(content)

            if is_folder:
                validated_paths.append(f'r"{content}"')
            else:
                # Files: resolve and validate existence
                try:
                    resolved_path = resolve_path(cleaned_path, must_exist=True)
                    validated_paths.append(resolved_path)
                except ValueError as e:
                    raise ValueError(f"Invalid file path '{path}': {e}") from e

        # remove dupes (but keep order)
        validated_paths = list(dict.fromkeys(validated_paths))
        self.docs_path = [path for path in validated_paths if path]

    @model_validator(mode="after")
    def validate_rag_user_data(self) -> Self:
        """Validate the RAG user data.

        Raises
        ------
        ValueError
            If the validation fails.

        Returns
        -------
        WaldiezRagUserProxyData
            The validated RAG user data.
        """
        self.validate_custom_embedding_function()
        self.validate_custom_token_count_function()
        self.validate_custom_text_split_function()
        self.validate_docs_path()
        if not self.db_config.model:
            self.db_config.model = WaldiezRagUserProxyModels[self.vector_db]
        if isinstance(self.n_results, int) and self.n_results < 1:
            self.n_results = None
        return self


def extract_raw_string_content(path: str) -> str:
    """Extract content from potential raw string formats.

    Parameters
    ----------
    path : str
        The path that might be wrapped in raw string format.

    Returns
    -------
    str
        The actual content of the path, without raw string formatting.
    """
    # Handle r"..." and r'...'
    if path.startswith(('r"', "r'")) and len(path) > 3:
        quote = path[1]
        if path.endswith(quote):
            return path[2:-1]
        # Handle malformed raw strings (missing end quote)
        return path[2:]
    return path


def string_represents_folder(path: str) -> bool:
    """Check if a string represents a folder.

    Parameters
    ----------
    path : str
        The string to check (does not need to exist).

    Returns
    -------
    bool
        True if the path is likely a folder, False if it's likely a file.
    """
    # Extract actual path content if wrapped
    content = extract_raw_string_content(path)

    # Explicit folder indicators
    if content.endswith(("/", "\\", os.path.sep)):
        return True

    # Check if it actually exists and is a directory
    try:
        if os.path.isdir(content):
            return True
    except (OSError, ValueError):  # pragma: no cover
        pass

    # Heuristic: no file extension likely means folder
    # return not os.path.splitext(content)[1]
    _, ext = os.path.splitext(path.rstrip("/\\"))
    return not ext


def is_remote_path(path: str) -> bool:
    """Check if a path is a remote path.

    Parameters
    ----------
    path : str
        The path to check.

    Returns
    -------
    tuple[bool, bool]
        If the path is a remote path and if it's a raw string.
    """
    content = extract_raw_string_content(path)
    for not_local in NOT_LOCAL:
        if content.startswith((not_local, f'r"{not_local}', f"r'{not_local}")):
            return True
    return False


def remove_file_scheme(path: str) -> str:
    """Remove the file:// scheme from a path.

    Parameters
    ----------
    path : str
        The path to remove the scheme from.

    Returns
    -------
    str
        The path without the scheme.
    """
    content = extract_raw_string_content(path)

    # Remove file:// prefix
    while content.startswith("file://"):
        content = content[len("file://") :]

    return f'r"{content}"'


def resolve_path(path: str, must_exist: bool) -> str:
    """Try to resolve a path.

    Parameters
    ----------
    path : str
        The path to resolve.
    must_exist : bool
        If the path must exist.

    Returns
    -------
    str
        The resolved path, potentially wrapped in raw string format.

    Raises
    ------
    ValueError
        If the path is not a valid local path.
    """
    # Extract the actual path content
    # if is_raw:
    path_content = extract_raw_string_content(path)
    # else:
    #     path_content = path

    # Handle JSON-escaped backslashes
    if "\\\\" in path_content:  # pragma: no cover
        path_content = path_content.replace("\\\\", "\\")
    # pylint: disable=too-many-try-statements
    try:
        # Try to resolve the path
        resolved = Path(path_content).resolve()

        if must_exist and not resolved.exists():
            raise ValueError(f"Path {path} does not exist.")

        return f'r"{resolved}"'

    except (
        OSError,
        UnicodeDecodeError,
        ValueError,
    ) as error:  # pragma: no cover
        # Fallback: try as raw string for Windows compatibility
        raw_version = f'r"{path_content}"'
        try:
            # Test if the path can be resolved when treated as raw
            resolved = Path(raw_version).resolve()
            if must_exist and not resolved.exists():
                raise ValueError(f"Path {path} does not exist.") from error
            return raw_version
        except Exception:
            raise ValueError(
                f"Path {path} is not a valid local path: {error}"
            ) from error



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Core exporting infrastructure.

This module provides base classes, types, result builders,
and extras for exportign flows.
"""

# Core base class
# Context management
# Constants / defaults
from .constants import (
    DEFAULT_AGENT_POSITION,
    DEFAULT_EXPORT_POSITION,
    DEFAULT_IMPORT_POSITION,
    FILE_HEADER,
)

# Content
from .content import ContentMetadata, PositionedContent

# Context management
from .context import (
    ExporterContext,
    create_exporter_context,
    get_default_exporter_context,
)

# Enums
from .enums import (
    AgentPosition,
    ContentOrder,
    ContentType,
    ExportPosition,
    GroupManagerStrategy,
    ImportPosition,
)

# Errors
from .errors import (
    ExporterContentError,
    ExporterError,
    ExporterInitializationError,
    ExporterValidationError,
)
from .exporter import (
    Exporter,
)
from .exporters import (
    ConfigurableExporter,
    SimpleExporter,
)

# Agent-specific extras
from .extras.agent_extras import (
    CaptainExtras,
    CodeExecutionConfig,
    GroupManagerExtras,
    RAGUserExtras,
    ReasoningExtras,
    StandardExtras,
    SystemMessageConfig,
    TerminationConfig,
)

# Base extras system
from .extras.base import BaseExtras

# Chat-specific extras
from .extras.chat_extras import (
    ChatExtras,
)

# Model-specific extras
from .extras.model_extras import (
    ModelExtras,
)

# Defaults
from .extras.path_resolver import DefaultPathResolver
from .extras.serializer import DefaultSerializer

# Tool-specific extras
from .extras.tool_extras import (
    ToolExtras,
)

# Protocols
from .protocols import (
    ContentGenerator,
    ExportContributor,
    ExportingLogger,
    PathResolver,
    Serializer,
    Validator,
)

# Export result system
from .result import (
    ExportResult,
    ExportResultBuilder,
    create_empty_result,
    create_result_with_content,
    merge_export_results,
)
from .types import (
    EnvironmentVariable,
    ExportConfig,
    Extras,
    ImportStatement,
    InstanceArgument,
    NoExtras,
)
from .utils import (
    get_agent_llm_config_arg,
    get_comment,
)
from .validation import (
    ValidationError,
    ValidationResult,
)

__all__ = [
    # Core base classes
    "Exporter",
    "SimpleExporter",
    "ConfigurableExporter",
    # Protocols
    "ContentGenerator",
    "ExportingLogger",
    "ExportContributor",
    "Validator",
    "Serializer",
    "PathResolver",
    # Core types and enums
    "ImportPosition",
    "ExportPosition",
    "AgentPosition",
    "ContentOrder",
    "GroupManagerStrategy",
    "Extras",
    "ImportStatement",
    "PositionedContent",
    "EnvironmentVariable",
    "ExportConfig",
    "ContentType",
    "ContentMetadata",
    "InstanceArgument",
    "NoExtras",
    "ValidationError",
    "ValidationResult",
    # Constants
    "FILE_HEADER",
    "DEFAULT_IMPORT_POSITION",
    "DEFAULT_EXPORT_POSITION",
    "DEFAULT_AGENT_POSITION",
    # Context management
    "ExporterContext",
    "get_default_exporter_context",
    "create_exporter_context",
    "create_exporter_context",
    # Export result system
    "ExportResult",
    "ExportResultBuilder",
    "merge_export_results",
    "create_empty_result",
    "create_result_with_content",
    # Base extras
    "BaseExtras",
    "DefaultSerializer",
    "DefaultPathResolver",
    # Agent extras
    "StandardExtras",
    "GroupManagerExtras",
    "CaptainExtras",
    "CodeExecutionConfig",
    "RAGUserExtras",
    "ReasoningExtras",
    "SystemMessageConfig",
    "TerminationConfig",
    # Model extras
    "ModelExtras",
    # Tool extras
    "ToolExtras",
    # Chat extras
    "ChatExtras",
    # Errors
    "ExporterError",
    "ExporterInitializationError",
    "ExporterValidationError",
    "ExporterContentError",
    # utils
    "get_comment",
    "get_agent_llm_config_arg",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Group chat manager agent."""

import warnings
from typing import Literal

from pydantic import Field, field_validator
from typing_extensions import Annotated

from ..agent import WaldiezAgent
from .group_manager_data import WaldiezGroupManagerData
from .speakers import WaldiezGroupManagerSpeakers


class WaldiezGroupManager(WaldiezAgent):
    """Group chat manager agent.

    A `WaldiezAgent` with agent_type `manager`, `human_input_mode`: `"NEVER"`
    and chat group related config for the agent.
    Also see `WaldiezAgent`, `WaldiezGroupManagerData`, `WaldiezAgentData`

    Attributes
    ----------
    agent_type : Literal["group_manager"]
        The agent type: 'group_manager' for a group manager agent
    data : WaldiezGroupManagerData
        The group manager agent's data.

    Functions
    ---------
    validate_transitions(agent_ids: list[str])
        Validate the transitions.
    """

    agent_type: Annotated[  # pyright: ignore
        Literal["group_manager", "manager"],
        Field(
            "manager",
            title="Agent type",
            description=(
                "The agent type: 'group_manager' for a group manager agent"
            ),
            alias="agentType",
        ),
    ] = "group_manager"
    data: Annotated[  # pyright: ignore
        WaldiezGroupManagerData,
        Field(
            title="Data",
            description="The group manager agent's data",
            default_factory=WaldiezGroupManagerData,  # pyright: ignore
        ),
    ]

    @field_validator("agent_type", mode="before")
    @classmethod
    def validate_agent_type(cls, v: str) -> Literal["group_manager"]:
        """Validate the agent type.

        The agent type must be `group_manager`.
        The other two are deprecated and will be removed in the future.

        Parameters
        ----------
        v : str
            The agent type.

        Returns
        -------
        str
            The validated agent type.

        Raises
        ------
        ValueError
            If the agent type is not `group_manager`.
        """
        if v in ["groupManager", "manager"]:  # pragma: no cover
            # Deprecated agent type names
            warnings.warn(
                (
                    "The agent types 'groupManager' and 'manager' are "
                    "deprecated. Use 'group_manager' instead."
                ),
                DeprecationWarning,
                stacklevel=2,
            )
        if v != "group_manager":  # pragma: no cover
            raise ValueError(
                "The agent type must be 'group_manager'. "
                "Use 'group_manager' instead."
            )
        return "group_manager"

    def validate_initial_agent_id(self, all_agent_ids: list[str]) -> None:
        """Validate the initial agent ID.

        The initial agent ID must be in the list of agent IDs.

        Parameters
        ----------
        all_agent_ids : list[str]
            The list of agent IDs.

        Raises
        ------
        ValueError
            If the initial agent ID is not in the list of agent IDs.
        """
        initial_agent_id = self.data.initial_agent_id
        if initial_agent_id not in all_agent_ids:
            raise ValueError(
                f"Initial agent ID '{initial_agent_id}' "
                f"is not in the list of agent IDs: {all_agent_ids}"
            )

    def get_speakers_order(self) -> list[str]:
        """Get the order of the speakers.

        Returns
        -------
        list[str]
            The order of the speakers.

        Raises
        ------
        RuntimeError
            If the order is not set.
        """
        return self.data.speakers.get_order()

    def set_speakers_order(self, group_members: list[str]) -> None:
        """Set the order of the speakers.

        Parameters
        ----------
        group_members : list[str]
            The group members' IDs.
        """
        self.data.speakers.set_order(self.data.initial_agent_id, group_members)

    def validate_transitions(self, agent_ids: list[str]) -> None:
        """Validate the transitions.

        If the selection mode is `transition`:

        - if `allow_repeat` is a list of agent_ids,
                make sure these ids exist.
        - make sure the `allowed_or_disallowed_transitions` mapping
                has valid agent ids.

        Parameters
        ----------
        agent_ids : list[str]
            The list of agent IDs.

        Raises
        ------
        ValueError
            If the transitions are invalid.
        """
        speakers: WaldiezGroupManagerSpeakers = self.data.speakers
        if speakers.selection_mode != "transition":
            return
        allow_repeat = speakers.allow_repeat
        if isinstance(allow_repeat, list):
            for agent_id in allow_repeat:
                if agent_id not in agent_ids:
                    raise ValueError(f"Invalid agent id: {agent_id}")
        for (
            agent_id,
            transitions,
        ) in speakers.allowed_or_disallowed_transitions.items():
            if agent_id not in agent_ids:
                raise ValueError(f"Invalid agent id: {agent_id}")
            for agent_id in transitions:
                if agent_id not in agent_ids:
                    raise ValueError(f"Invalid agent id: {agent_id}")



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Assistant agent data model."""

from pydantic import Field
from typing_extensions import Annotated, Literal

from ..agent import WaldiezAgentData


class WaldiezAssistantData(WaldiezAgentData):
    """Assistant agent data class.

    The data for an agent with `human_input_mode` set to `"ALWAYS"` as default.
    See the parent's docs (`WaldiezAgentData`) for the rest of the properties.

    Attributes
    ----------
    human_input_mode : Literal["ALWAYS", "NEVER", "TERMINATE"]
        The human input mode, Defaults to `NEVER`
    is_multimodal : bool
        A flag to indicate if the agent is multimodal.
        Defaults to `False`.
    """

    human_input_mode: Annotated[
        Literal["ALWAYS", "NEVER", "TERMINATE"],
        Field(
            "NEVER",
            title="Human input mode",
            description="The human input mode, Defaults to `NEVER`",
            alias="humanInputMode",
        ),
    ] = "NEVER"
    is_multimodal: Annotated[
        bool,
        Field(
            False,
            title="Is multimodal",
            description="A flag to indicate if the agent is multimodal.",
            alias="isMultimodal",
        ),
    ] = False



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez Tool types."""

from typing_extensions import Literal

WaldiezToolType = Literal["shared", "custom", "langchain", "crewai"]
"""Possible types of a Waldiez Tool."""



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=no-self-use,too-few-public-methods
"""Group member agent configuration processor."""

from dataclasses import dataclass, field
from typing import Callable

from waldiez.models import (
    WaldiezAgent,
    WaldiezAgentUpdateSystemMessage,
    WaldiezChat,
)

from ...core import (
    ImportStatement,
    InstanceArgument,
)
from .handoffs import HandoffProcessor


@dataclass
class GroupMemberProcessorResult:
    """Result from processing group member agent configuration."""

    before_agent: str = ""
    after_agent: str = ""
    extra_arguments: list[InstanceArgument] = field(
        default_factory=list[InstanceArgument]
    )
    extra_imports: set[ImportStatement] = field(
        default_factory=set[ImportStatement]
    )


class GroupMemberAgentProcessor:
    """Processor for group member agent configuration."""

    result: GroupMemberProcessorResult

    def __init__(
        self,
        agent: WaldiezAgent,
        agent_names: dict[str, str],
        tool_names: dict[str, str],
        all_chats: list[WaldiezChat],
        chat_names: dict[str, str],
        serializer: Callable[..., str],
    ) -> None:
        self.agent = agent
        self.agent_names = agent_names
        self.tool_names = tool_names
        self.all_chats = all_chats
        self.chat_names = chat_names
        self.serializer = serializer

    def process(
        self,
    ) -> GroupMemberProcessorResult:
        """Process group member agent configuration.

        Returns
        -------
        GroupMemberProcessorResult
            The processed result containing before_agent code,
            after_agent code, and extra imports.
        """
        self.result = GroupMemberProcessorResult()
        if not self.agent.is_group_member:
            return self.result
        self._process_agent_functions()
        self._process_agent_update_state_before_reply()
        self._process_handoff_registrations()
        return self.result

    def _process_agent_functions(self) -> None:
        """Process function arguments for the agent."""
        tab = "    "
        arg_string = "["

        tools_added = self._add_tools()
        if tools_added:
            arg_string += f"{tools_added}\n{tab}"

        arg_string += "]"
        self.result.extra_arguments.append(
            InstanceArgument(
                name="functions",
                value=arg_string,
                instance_id=self.agent.id,
                tabs=1,
            )
        )

    def _add_tools(self) -> str:
        """Add tools to the function arguments."""
        tab = "    "
        arg_string = ""
        for tool in self.agent.data.tools:
            tool_name = self.tool_names.get(tool.id)
            if tool_name:
                arg_string += f"\n{tab}{tab}{tool_name},"
        return arg_string

    def _process_agent_update_state_before_reply(self) -> None:
        """Process agent update state before reply configuration."""
        tab = "    "
        arg_string = "["

        arcg_content = self._process_update_functions()

        if arcg_content:
            arg_string += f"{arcg_content}\n{tab}"
            self.result.extra_imports.add(
                ImportStatement("from autogen import UpdateSystemMessage")
            )
        arg_string += "]"
        self.result.extra_arguments.append(
            InstanceArgument(
                name="update_agent_state_before_reply",
                value=arg_string,
                instance_id=self.agent.id,
                tabs=1,
            )
        )

    def _process_update_functions(
        self,
    ) -> str:
        """Process individual update functions."""
        arg_content = ""

        for function in self.agent.data.update_agent_state_before_reply:
            if isinstance(function, WaldiezAgentUpdateSystemMessage):
                arg_content = self._process_system_message_update(function)
            else:
                arg_content = self._process_tool_update(function)

        return arg_content

    def _process_system_message_update(
        self,
        function: WaldiezAgentUpdateSystemMessage,
    ) -> str:
        """Process system message update function."""
        tab = "    "
        space = tab * 2
        if function.type == "callable":
            content, name = function.get_content(
                name_suffix=self.agent_names[self.agent.id]
            )
            arg_string = f"\n{space}UpdateSystemMessage({name}),"
            self.result.before_agent += content + "\n"
        else:
            dumped = self.serializer(function.content)
            arg_string = f"\n{space}UpdateSystemMessage({dumped}),"
        return arg_string

    def _process_tool_update(
        self,
        function_id: str,
    ) -> str:
        """Process tool-based update function."""
        tool_name = self.tool_names.get(function_id)
        if tool_name:
            tab = "    "
            return f"\n{tab}{tab}{tool_name},"
        return ""

    def _process_handoff_registrations(self) -> None:
        """Process handoff registrations for the agent."""
        result = HandoffProcessor(
            agent=self.agent,
            agent_names=self.agent_names,
            chat_names=self.chat_names,
            all_chats=self.all_chats,
            serializer=self.serializer,
        ).process()
        for extra_import in result.extra_imports:
            self.result.extra_imports.add(ImportStatement(extra_import))
        self.result.after_agent += result.after_agent



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Chat data model."""

from typing import Any, Optional, Union

from pydantic import Field, field_validator, model_validator
from typing_extensions import Annotated, Self

from ..agents import WaldiezAgentType
from ..common import (
    WaldiezBase,
    WaldiezDefaultCondition,
    WaldiezHandoffCondition,
    WaldiezHandoffTransition,
    WaldiezTransitionAvailability,
    check_function,
    update_dict,
)
from .chat_message import (
    CALLABLE_MESSAGE,
    CALLABLE_MESSAGE_ARGS,
    WaldiezChatMessage,
)
from .chat_nested import WaldiezChatNested
from .chat_summary import WaldiezChatSummary


class WaldiezChatData(WaldiezBase):
    """Chat data class.

    Attributes
    ----------
    name : str
        The name of the chat.
    source_type : WaldiezAgentType
        The agent type of the chat source.
    target_type : WaldiezAgentType
        The agent type of the chat target.
    description : str
        The description of the chat.
    position : int
        The position of the chat. Ignored (UI related).
    order : int
        The of the chat. If negative, ignored.
    clear_history : Optional[bool], optional
        Whether to clear the chat history, by default None.
    message : Union[str, WaldiezChatMessage]
        The message of the chat.
    nested_chat : WaldiezChatNested
        The nested chat config.
    summary : WaldiezChatSummary
        The summary method and options for the chat.
    max_turns : Optional[int]
        The maximum number of turns for the chat, by default None (no limit).
    silent : bool, optional
        Whether to run the chat silently, by default False (not silent).
    summary_args : Optional[dict[str, Any]]
        The summary args to use in autogen.
    handoff_condition : Optional[WaldiezHandoffCondition], optional
        The handoff condition to use, by default None (for group chat).
    real_source : Optional[str]
        The real source of the chat (overrides the source).
    real_target : Optional[str]
        The real target of the chat (overrides the target).
    """

    name: Annotated[
        str, Field(..., title="Name", description="The name of the chat.")
    ]
    description: Annotated[
        str,
        Field(
            default="A new chat",
            title="Description",
            description="The description of the chat.",
        ),
    ] = "A new chat"
    position: Annotated[
        int,
        Field(
            default=-1,
            title="Position",
            description="The position of the chat in the flow (Ignored).",
        ),
    ] = -1
    order: Annotated[
        int,
        Field(
            default=-1,
            title="Order",
            description="The order of the chat in the flow.",
        ),
    ] = -1
    clear_history: Annotated[
        bool,
        Field(
            default=True,
            alias="clearHistory",
            title="Clear History",
            description="Whether to clear the chat history.",
        ),
    ] = True
    message: Annotated[
        Union[str, WaldiezChatMessage],
        Field(
            title="Message",
            description="The message of the chat.",
            default_factory=WaldiezChatMessage,
        ),
    ]
    nested_chat: Annotated[
        WaldiezChatNested,
        Field(
            title="Nested Chat",
            description="The nested chat.",
            alias="nestedChat",
            default_factory=WaldiezChatNested,
        ),
    ]
    summary: Annotated[
        WaldiezChatSummary,
        Field(
            default_factory=WaldiezChatSummary,
            title="Summary",
            description="The summary method options for the chat.",
        ),
    ]
    max_turns: Annotated[
        Optional[int],
        Field(
            default=None,
            alias="maxTurns",
            title="Max Turns",
            description="The maximum number of turns for the chat.",
        ),
    ] = None
    prerequisites: Annotated[
        list[str],
        Field(
            title="Prerequisites",
            description="The prerequisites (chat ids) for the chat (if async).",
            default_factory=list,
        ),
    ] = []
    silent: Annotated[
        bool,
        Field(
            default=False,
            title="Silent",
            description="Whether to run the chat silently.",
        ),
    ] = False
    real_source: Annotated[
        Optional[str],
        Field(
            default=None,
            alias="realSource",
            title="Real Source",
            description="The real source of the chat (overrides the source).",
        ),
    ] = None
    real_target: Annotated[
        Optional[str],
        Field(
            default=None,
            alias="realTarget",
            title="Real Target",
            description="The real target of the chat (overrides the target).",
        ),
    ] = None
    source_type: Annotated[
        WaldiezAgentType,
        Field(
            ...,
            alias="sourceType",
            title="Source Type",
            description="The agent type of the source.",
        ),
    ]
    target_type: Annotated[
        WaldiezAgentType,
        Field(
            ...,
            alias="targetType",
            title="Target Type",
            description="The agent type of the target.",
        ),
    ]
    condition: Annotated[
        WaldiezHandoffCondition,
        Field(
            default_factory=WaldiezDefaultCondition.create,
            alias="condition",
            title="Handoff Condition",
            description="The handoff condition to use.",
        ),
    ]
    available: Annotated[
        WaldiezTransitionAvailability,
        Field(
            default_factory=WaldiezTransitionAvailability,
            title="Availability",
            description="The availability condition for the chat.",
        ),
    ]
    after_work: Annotated[
        Optional[WaldiezHandoffTransition],
        Field(
            None,
            title="After Work",
            description=(
                "The target to transfer control to after the chat has "
                "finished its work. (used if in a group chat)"
            ),
            alias="afterWork",
        ),
    ] = None
    _message_content: Optional[str] = None
    _chat_id: int = 0
    _prerequisites: list[int] = []

    @property
    def message_content(self) -> Optional[str]:
        """Get the message content."""
        return self._message_content

    def get_chat_id(self) -> int:
        """Get the chat id.

        Returns
        -------
        int
            The chat id.
        """
        return self._chat_id

    def set_chat_id(self, value: int) -> None:
        """Set the chat id.

        Parameters
        ----------
        value : int
            The chat id.
        """
        self._chat_id = value

    def get_prerequisites(self) -> list[int]:
        """Get the chat prerequisites.

        Returns
        -------
        list[int]
            The chat prerequisites (if async).
        """
        return self._prerequisites

    def set_prerequisites(self, value: list[int]) -> None:
        """Set the chat prerequisites.

        Parameters
        ----------
        value : list[int]
            The chat prerequisites to set.
        """
        self._prerequisites = value

    @model_validator(mode="after")
    def validate_chat_data(self) -> Self:
        """Validate the chat data.

        Returns
        -------
        WaldiezChatData
            The validated chat data.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if not isinstance(self.message, WaldiezChatMessage):  # pragma: no cover
            return self
        self._message_content = self.message.content
        if self.message.type == "none":
            self._message_content = None
        if self.message.type == "string":
            self._message_content = self.message.content
        if self.message.type == "method":
            valid, error_or_body = check_function(
                self.message.content or "",
                CALLABLE_MESSAGE,
                CALLABLE_MESSAGE_ARGS,
            )
            if not valid:
                raise ValueError(error_or_body)
            self._message_content = error_or_body
        return self

    # noinspection PyNestedDecorators
    @field_validator("message", mode="before")
    @classmethod
    def validate_message(cls, value: Any) -> WaldiezChatMessage:
        """Validate the message.

        Parameters
        ----------
        value : Any
            The message value.

        Returns
        -------
        WaldiezChatMessage
            The validated message value.

        Raises
        ------
        ValueError
            If the validation fails.
        """
        if value is None:
            return WaldiezChatMessage(
                type="none", use_carryover=False, content=None, context={}
            )
        if isinstance(value, (str, int, float, bool)):
            return WaldiezChatMessage(
                type="string",
                use_carryover=False,
                content=str(value),
                context={},
            )
        if isinstance(value, dict):
            return WaldiezChatMessage.model_validate(value)
        if not isinstance(value, WaldiezChatMessage):
            return WaldiezChatMessage(
                type="none", use_carryover=False, content=None, context={}
            )
        return value

    @property
    def summary_args(self) -> Optional[dict[str, Any]]:
        """Get the summary args."""
        if self.summary.method not in (
            "reflection_with_llm",
            "reflectionWithLlm",
        ):
            return None
        args: dict[str, Any] = {}
        if self.summary.prompt:  # pragma: no branch
            args["summary_prompt"] = self.summary.prompt
        if self.summary.args:  # pragma: no branch
            args.update(self.summary.args)
        return args

    def _get_context_args(self) -> dict[str, Any]:
        """Get the context arguments to use in autogen.

        Returns
        -------
        dict[str, Any]
            The dictionary to use for generating the kwargs.
        """
        extra_args: dict[str, Any] = {}
        if not isinstance(self.message, WaldiezChatMessage):  # pragma: no cover
            return extra_args
        extra_args.update(update_dict(self.message.context))
        return extra_args

    def get_chat_args(self, for_queue: bool) -> dict[str, Any]:
        """Get the chat arguments to use in autogen.

        Without the 'message' key.

        Parameters
        ----------
        for_queue : bool
            Whether to get the arguments for a chat queue.

        Returns
        -------
        dict[str, Any]
            The dictionary to pass as kwargs.
        """
        args: dict[str, Any] = {}
        if self.summary.method:  # pragma: no branch
            args["summary_method"] = str(self.summary.method)
        if self.summary_args:
            args["summary_args"] = self.summary_args
        if (
            isinstance(self.max_turns, int) and self.max_turns > 0
        ):  # pragma: no branch
            args["max_turns"] = self.max_turns
        args["clear_history"] = self.clear_history
        # args["silent"] = self.silent
        args.update(self._get_context_args())
        if for_queue:
            args["chat_id"] = self._chat_id
        if self._prerequisites:
            args["prerequisites"] = self._prerequisites
        return args



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Agent models."""

from .agent import (
    IS_TERMINATION_MESSAGE,
    IS_TERMINATION_MESSAGE_ARGS,
    IS_TERMINATION_MESSAGE_TYPES,
    WaldiezAgent,
    WaldiezAgentCodeExecutionConfig,
    WaldiezAgentData,
    WaldiezAgentHumanInputMode,
    WaldiezAgentLinkedTool,
    WaldiezAgentNestedChat,
    WaldiezAgentNestedChatMessage,
    WaldiezAgentTerminationMessage,
    WaldiezAgentType,
    WaldiezAgentUpdateFunctionType,
    WaldiezAgentUpdateSystemMessage,
)
from .agents import WaldiezAgents
from .assistant import WaldiezAssistant, WaldiezAssistantData
from .captain import (
    WaldiezCaptainAgent,
    WaldiezCaptainAgentData,
    WaldiezCaptainAgentLibEntry,
)
from .extra_requirements import (
    get_captain_agent_extra_requirements,
    get_retrievechat_extra_requirements,
)
from .group_manager import (
    CUSTOM_SPEAKER_SELECTION,
    CUSTOM_SPEAKER_SELECTION_ARGS,
    CUSTOM_SPEAKER_SELECTION_TYPES,
    WaldiezGroupManager,
    WaldiezGroupManagerData,
    WaldiezGroupManagerSpeakers,
    WaldiezGroupManagerSpeakersSelectionMethod,
    WaldiezGroupManagerSpeakersSelectionMode,
    WaldiezGroupManagerSpeakersTransitionsType,
)
from .rag_user_proxy import (
    CUSTOM_EMBEDDING_FUNCTION,
    CUSTOM_EMBEDDING_FUNCTION_ARGS,
    CUSTOM_EMBEDDING_FUNCTION_TYPES,
    CUSTOM_TEXT_SPLIT_FUNCTION,
    CUSTOM_TEXT_SPLIT_FUNCTION_ARGS,
    CUSTOM_TEXT_SPLIT_FUNCTION_TYPES,
    CUSTOM_TOKEN_COUNT_FUNCTION,
    CUSTOM_TOKEN_COUNT_FUNCTION_ARGS,
    CUSTOM_TOKEN_COUNT_FUNCTION_TYPES,
    WaldiezRagUserProxy,
    WaldiezRagUserProxyChunkMode,
    WaldiezRagUserProxyData,
    WaldiezRagUserProxyModels,
    WaldiezRagUserProxyRetrieveConfig,
    WaldiezRagUserProxyTask,
    WaldiezRagUserProxyVectorDb,
    WaldiezRagUserProxyVectorDbConfig,
)
from .reasoning import (
    WaldiezReasoningAgent,
    WaldiezReasoningAgentData,
    WaldiezReasoningAgentReasonConfig,
)
from .user_proxy import WaldiezUserProxy, WaldiezUserProxyData

__all__ = [
    "get_retrievechat_extra_requirements",
    "get_captain_agent_extra_requirements",
    "IS_TERMINATION_MESSAGE",
    "IS_TERMINATION_MESSAGE_ARGS",
    "IS_TERMINATION_MESSAGE_TYPES",
    "CUSTOM_EMBEDDING_FUNCTION",
    "CUSTOM_EMBEDDING_FUNCTION_ARGS",
    "CUSTOM_EMBEDDING_FUNCTION_TYPES",
    "CUSTOM_TEXT_SPLIT_FUNCTION",
    "CUSTOM_TEXT_SPLIT_FUNCTION_ARGS",
    "CUSTOM_TEXT_SPLIT_FUNCTION_TYPES",
    "CUSTOM_TOKEN_COUNT_FUNCTION",
    "CUSTOM_TOKEN_COUNT_FUNCTION_ARGS",
    "CUSTOM_TOKEN_COUNT_FUNCTION_TYPES",
    "CUSTOM_SPEAKER_SELECTION",
    "CUSTOM_SPEAKER_SELECTION_ARGS",
    "CUSTOM_SPEAKER_SELECTION_TYPES",
    "WaldiezAgent",
    "WaldiezAgentType",
    "WaldiezAgents",
    "WaldiezAssistant",
    "WaldiezAssistantData",
    "WaldiezAgentCodeExecutionConfig",
    "WaldiezAgentData",
    "WaldiezAgentHumanInputMode",
    "WaldiezAgentLinkedTool",
    "WaldiezAgentNestedChat",
    "WaldiezAgentNestedChatMessage",
    "WaldiezAgentTerminationMessage",
    "WaldiezCaptainAgent",
    "WaldiezCaptainAgentData",
    "WaldiezCaptainAgentLibEntry",
    "WaldiezRagUserProxy",
    "WaldiezRagUserProxyData",
    "WaldiezRagUserProxyModels",
    "WaldiezReasoningAgent",
    "WaldiezReasoningAgentData",
    "WaldiezReasoningAgentReasonConfig",
    "WaldiezUserProxy",
    "WaldiezUserProxyData",
    "WaldiezRagUserProxyRetrieveConfig",
    "WaldiezRagUserProxyTask",
    "WaldiezRagUserProxyChunkMode",
    "WaldiezRagUserProxyVectorDb",
    "WaldiezRagUserProxyVectorDbConfig",
    "WaldiezGroupManager",
    "WaldiezGroupManagerData",
    "WaldiezGroupManagerSpeakers",
    "WaldiezGroupManagerSpeakersSelectionMethod",
    "WaldiezGroupManagerSpeakersSelectionMode",
    "WaldiezGroupManagerSpeakersTransitionsType",
    "WaldiezAgentUpdateSystemMessage",
    "WaldiezAgentUpdateFunctionType",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Extras for Waldiez exporting core module."""

from .agent_extras import (
    CaptainExtras,
    CodeExecutionConfig,
    GroupManagerExtras,
    RAGUserExtras,
    ReasoningExtras,
    StandardExtras,
    SystemMessageConfig,
    TerminationConfig,
)
from .base import BaseExtras
from .chat_extras import ChatExtras
from .flow_extras import FlowExtras
from .model_extras import ModelExtras
from .path_resolver import DefaultPathResolver
from .serializer import DefaultSerializer
from .tool_extras import ToolExtras

__all__ = [
    "BaseExtras",
    "ChatExtras",
    "FlowExtras",
    "ModelExtras",
    "DefaultPathResolver",
    "DefaultSerializer",
    "ToolExtras",
    "StandardExtras",
    "CaptainExtras",
    "CodeExecutionConfig",
    "SystemMessageConfig",
    "TerminationConfig",
    "RAGUserExtras",
    "ReasoningExtras",
    "GroupManagerExtras",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods
"""Agent specific extras module."""

from .captain_extras import CaptainExtras
from .group_manager_extras import GroupManagerExtras
from .rag_user_extras import RAGUserExtras
from .reasoning_extras import ReasoningExtras
from .standard_extras import (
    CodeExecutionConfig,
    StandardExtras,
    SystemMessageConfig,
    TerminationConfig,
)

__all__ = [
    "StandardExtras",
    "CaptainExtras",
    "StandardExtras",
    "RAGUserExtras",
    "ReasoningExtras",
    "GroupManagerExtras",
    "CodeExecutionConfig",
    "SystemMessageConfig",
    "TerminationConfig",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods
"""Termination message configuration for Waldiez agents."""

from waldiez.models import WaldiezAgent

from ..core import TerminationConfig


class TerminationProcessor:
    """Processor for termination message configuration."""

    def __init__(
        self,
        agent: WaldiezAgent,
        agent_name: str,
    ):
        self.agent = agent
        self.agent_name = agent_name

    def process(self) -> TerminationConfig:
        """Process termination configuration.

        Returns
        -------
        TerminationConfig
            The processed termination configuration.
        """
        if self.agent.data.termination.type == "none":
            return TerminationConfig(
                termination_arg="None",
                before_content="",
            )
        if self.agent.data.termination.type == "keyword":
            return TerminationConfig(
                termination_arg=self.agent.data.termination.string,
                before_content="",
            )
        if self.agent.data.termination.type == "method":
            content, function_name = (
                self.agent.data.termination.get_termination_function(
                    name_suffix=self.agent_name
                )
            )
            return TerminationConfig(
                termination_arg=function_name,
                before_content="\n\n" + content + "\n",
            )
        return TerminationConfig()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez model."""

from .extra_requirements import get_models_extra_requirements
from .model import DEFAULT_BASE_URLS, MODEL_NEEDS_BASE_URL, WaldiezModel
from .model_data import (
    WaldiezModelAPIType,
    WaldiezModelAWS,
    WaldiezModelData,
    WaldiezModelPrice,
)

__all__ = [
    "get_models_extra_requirements",
    "DEFAULT_BASE_URLS",
    "MODEL_NEEDS_BASE_URL",
    "WaldiezModel",
    "WaldiezModelData",
    "WaldiezModelPrice",
    "WaldiezModelAPIType",
    "WaldiezModelAWS",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Try to suppress the warning about flaml.automl not being available."""

import logging

__waldiez_checked_flaml_warnings = False  # pylint: disable=invalid-name


def check_flaml_warnings() -> None:  # pragma: no cover
    """Check for flaml warnings once."""
    # pylint: disable=global-statement
    global __waldiez_checked_flaml_warnings
    if __waldiez_checked_flaml_warnings is False:
        flam_logger = logging.getLogger("flaml")
        flam_logger.setLevel(logging.ERROR)
        __waldiez_checked_flaml_warnings = True



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods
"""Handoff processor for Waldiez agents."""

from dataclasses import dataclass, field
from typing import Callable

from waldiez.models import (
    WaldiezAgent,
    WaldiezChat,
    WaldiezHandoff,
    WaldiezTransitionTarget,
)

from .available import (
    TransitionAvailableProcessor,
    TransitionAvailableResult,
)
from .condition import ConditionProcessor, ConditionResult
from .target import TargetResult, TransitionTargetProcessor


@dataclass
class HandoffResult:
    """Result from processing handoffs.

    Attributes
    ----------
        Content to be placed before the agent definition.
    after_agent : str
        Content to be placed after the agent definition.
    extra_imports : set[str]
        Additional imports required for the handoff.
    """

    after_agent: str = ""
    extra_imports: set[str] = field(default_factory=set[str])


class HandoffProcessor:
    """Processor for agent handoffs."""

    result: HandoffResult
    IMPORT_PREFIX = "from autogen.agentchat.group import "

    def __init__(
        self,
        agent: WaldiezAgent,
        agent_names: dict[str, str],
        chat_names: dict[str, str],
        all_chats: list[WaldiezChat],
        serializer: Callable[..., str],
    ) -> None:
        self.agent = agent
        self.agent_names = agent_names
        self.agent_name = agent_names[agent.id]
        self.chat_names = chat_names
        self.all_chats = all_chats
        self.serializer = serializer
        self.result = HandoffResult()

    def process(self) -> HandoffResult:
        """Process all handoff configurations.

        Returns
        -------
        HandoffResult
            The processed handoff result containing after_agent code,
            and extra imports.
        """
        for handoff in self.agent.handoffs:
            if handoff.is_empty():
                continue
            target = TransitionTargetProcessor(
                agent=self.agent,
                agent_names=self.agent_names,
                chat_names=self.chat_names,
                all_chats=self.all_chats,
                serializer=self.serializer,
            ).process(target=handoff.target)
            self.result.extra_imports.update(target.extra_imports)
            if target.before_content:
                self.result.after_agent += f"{target.before_content}\n"
            condition = ConditionProcessor(
                agent=self.agent,
                agent_names=self.agent_names,
                chat_names=self.chat_names,
                all_chats=self.all_chats,
                serializer=self.serializer,
            ).process(condition=handoff.condition)
            self.result.extra_imports.update(condition.extra_imports)
            available = TransitionAvailableProcessor(
                serializer=self.serializer,
            ).process(
                available=handoff.available,
            )
            registration = self._create_handoff_registration(
                handoff=handoff,
                target=target,
                condition=condition,
                available=available,
            )
            self.result.extra_imports.update(available.extra_imports)
            if registration:
                self.result.after_agent += f"{registration}\n"

        if self.agent.data.after_work:
            after_work_result = self._process_after_work(
                self.agent.data.after_work
            )
            if after_work_result.before_content:
                self.result.after_agent += (
                    f"{after_work_result.before_content}\n"
                )
            self.result.after_agent += after_work_result.content
            self.result.extra_imports.update(after_work_result.extra_imports)
        return self.result

    def _create_handoff_registration(
        self,
        handoff: WaldiezHandoff,
        target: TargetResult,
        condition: ConditionResult,
        available: TransitionAvailableResult,
    ) -> str:
        """Create the handoff registration string."""
        if handoff.is_empty():
            return ""
        reg_string = ""
        if handoff.is_llm_based():
            reg_string += f"{self.agent_name}.handoffs.add_llm_condition(\n"
            reg_string += "    condition=OnCondition(\n"
            self.result.extra_imports.add(f"{self.IMPORT_PREFIX}OnCondition")
        else:
            reg_string += f"{self.agent_name}.handoffs.add_context_condition(\n"
            reg_string += "    condition=OnContextCondition(\n"
            self.result.extra_imports.add(
                f"{self.IMPORT_PREFIX}OnContextCondition"
            )
        reg_string += f"        target={target.content},\n"
        reg_string += f"        condition={condition.content},\n"
        if available.content:
            reg_string += f"        available={available.content},\n"
        reg_string += "    )\n"
        reg_string += ")"
        return reg_string

    def _process_after_work(
        self, after_work: WaldiezTransitionTarget
    ) -> TargetResult:
        target_processor = TransitionTargetProcessor(
            agent=self.agent,
            agent_names=self.agent_names,
            chat_names=self.chat_names,
            all_chats=self.all_chats,
            serializer=self.serializer,
        )
        target_result = target_processor.process(after_work)
        if not target_result.content:
            return TargetResult()
        registration = (
            f"{self.agent_name}.handoffs.set_after_work(\n"
            f"    target={target_result.content}\n"
            ")"
        )
        return TargetResult(
            content=registration,
            before_content=target_result.before_content,
            extra_imports=target_result.extra_imports,
        )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=no-self-use,too-few-public-methods
"""Group manager agent configuration processor."""

from typing import Callable, Optional

from waldiez.models import (
    WaldiezAgent,
    WaldiezAgentConnection,
    WaldiezGroupManager,
    WaldiezModel,
    WaldiezTransitionTarget,
)

from ...core import (
    CodeExecutionConfig,
    GroupManagerExtras,
    GroupManagerStrategy,
    SystemMessageConfig,
    TerminationConfig,
    get_agent_llm_config_arg,
)


class GroupManagerProcessor:
    """Processes group manager configurations."""

    def __init__(
        self,
        agent: WaldiezAgent,
        initial_chats: list[WaldiezAgentConnection],
        group_chat_members: list[WaldiezAgent],
        agent_names: dict[str, str],
        model_names: dict[str, str],
        all_models: list[WaldiezModel],
        serializer: Callable[..., str],
        cache_seed: Optional[int] = None,
    ):
        """Initialize the group manager processor.

        Parameters
        ----------
        agent : WaldiezAgent
            The group manager agent to process.
        initial_chats : list[WaldiezAgentConnection]
            The initial chats to process for the group manager.
        group_chat_members : list[WaldiezAgent]
            The members of the group chat.
        agent_names : dict[str, str]
            Mapping of agent IDs to their names.
        model_names : dict[str, str]
            Mapping of model IDs to their names.
        all_models : list[WaldiezModel]
            List of all models available for the agent.
        serializer : Callable[..., str]
            Function to serialize data into a string format.
        cache_seed : Optional[int], optional
            Seed for caching purposes, by default None.
        """
        self.agent = agent
        self.initial_chats = initial_chats
        self.group_chat_members = group_chat_members
        self.agent_names = agent_names
        self.model_names = model_names
        self.all_models = all_models
        self.serializer = serializer
        self.cache_seed = cache_seed

    def process(
        self,
        code_execution_config: Optional[CodeExecutionConfig] = None,
        termination_config: Optional[TerminationConfig] = None,
        system_message_config: Optional[SystemMessageConfig] = None,
    ) -> GroupManagerExtras:
        """Process group manager and return extras.

        Parameters
        ----------
        code_execution_config : CodeExecutionConfig, optional
            Configuration for code execution, by default None.
        termination_config : TerminationConfig, optional
            Configuration for termination, by default None.
        system_message_config : SystemMessageConfig, optional
            Configuration for system messages, by default None.

        Returns
        -------
        GroupManagerExtras
            The processed group manager extras containing configuration.
        """
        extras = GroupManagerExtras(
            instance_id=self.agent.id,
            code_execution_config=code_execution_config,
            termination_config=termination_config,
            system_message_config=system_message_config,
        )

        # Determine strategy based on initial chats and message types
        extras.strategy = self._determine_strategy()

        if extras.strategy == GroupManagerStrategy.PATTERN:
            self._process_pattern_strategy(extras)
        else:
            self._process_traditional_strategy(extras)

        return extras

    @staticmethod
    def is_pattern_strategy(
        initial_chats: list[WaldiezAgentConnection],
    ) -> bool:
        """Check if the group manager should use pattern strategy.

        Parameters
        ----------
        initial_chats : list[WaldiezAgentConnection]
            The initial chats to determine the strategy.

        Returns
        -------
        bool
            True if pattern strategy should be used, False otherwise.
        """
        if not initial_chats:
            return True

        first_chat = initial_chats[0]["chat"]
        return (
            isinstance(first_chat.data.message, str)
            or not first_chat.data.message.is_method()
        )

    def _determine_strategy(self) -> GroupManagerStrategy:
        """Determine which strategy to use for this group manager."""
        if self.is_pattern_strategy(self.initial_chats):
            # Use AG2 Pattern system
            return GroupManagerStrategy.PATTERN

        # Method call - must use traditional approach
        return GroupManagerStrategy.TRADITIONAL

    def _process_pattern_strategy(self, extras: GroupManagerExtras) -> None:
        """Process using AG2 Pattern system."""
        extras.pattern_class_name = self._get_pattern_class_name()

        # Add required imports
        extras.pattern_imports.add(
            "from autogen.agentchat.group.patterns import "
            f"{extras.pattern_class_name}"
        )

        # Get user agent if applicable
        user_agent = None
        if self.initial_chats:
            user_agent = self.initial_chats[0]["source"]

        # Generate pattern definition
        extras.pattern_definition = self._generate_pattern_definition(
            extras.pattern_class_name, user_agent
        )

        # Add context variables if present
        if self.agent.data.context_variables:
            ctx_lines = self._generate_context_variables()
            if ctx_lines:
                extras.context_variables_content = (
                    "    context_variables=ContextVariables(data={\n"
                    + "\n".join(ctx_lines)
                    + "\n    })"
                )
            extras.pattern_imports.add(
                "from autogen.agentchat.group import ContextVariables"
            )

        # Add after work configuration if needed
        if self._should_check_for_after_work(extras.pattern_class_name):
            after_work_arg, extra_import = self._get_after_work_configuration()
            if after_work_arg:
                extras.after_work_content = after_work_arg
            if extra_import:
                extras.pattern_imports.add(extra_import)

    def _process_traditional_strategy(self, extras: GroupManagerExtras) -> None:
        """Process using traditional GroupChat + GroupChatManager."""
        if not self.initial_chats:
            return

        user_agent = self.initial_chats[0]["source"]
        group_members = self.group_chat_members
        if user_agent not in group_members:
            group_members.append(user_agent)
        admin_name = self.agent_names[user_agent.id]

        # Generate group chat definition
        group_chat_content = self._generate_group_chat_definition(
            group_members, admin_name
        )

        extras.group_chat_definition = group_chat_content["definition"]
        extras.group_chat_name = group_chat_content["name"]
        extras.custom_speaker_selection = group_chat_content["custom_selection"]

        # Generate group chat argument for agent constructor
        if extras.group_chat_name:
            extras.group_chat_argument = f"groupchat={extras.group_chat_name}"

    def _get_pattern_class_name(self) -> str:
        """Get the appropriate pattern class name."""
        if not isinstance(self.agent, WaldiezGroupManager):
            return "AutoPattern"
        selection_method = self.agent.data.speakers.selection_method

        pattern_map = {
            "auto": "AutoPattern",
            "manual": "ManualPattern",
            "round_robin": "RoundRobinPattern",
            "random": "RandomPattern",
            "default": "DefaultPattern",
        }

        return pattern_map.get(selection_method, "AutoPattern")

    def _generate_pattern_definition(
        self, pattern_class: str, user_agent: Optional[WaldiezAgent]
    ) -> str:
        """Generate the pattern definition string."""
        if not isinstance(self.agent, WaldiezGroupManager):
            return ""
        manager_name = self.agent_names[self.agent.id]
        initial_agent_name = self.agent_names[self.agent.data.initial_agent_id]
        agents_string = ", ".join(
            self.agent_names[agent.id] for agent in self.group_chat_members
        )

        pattern_lines = [
            f"{manager_name}_pattern = {pattern_class}(",
            f"    initial_agent={initial_agent_name},",
            f"    agents=[{agents_string}],",
        ]

        if user_agent:
            pattern_lines.append(
                f"    user_agent={self.agent_names[user_agent.id]},"
            )

        # Add LLM config
        llm_config_arg = get_agent_llm_config_arg(
            agent=self.agent,
            all_models=self.all_models,
            model_names=self.model_names,
            cache_seed=self.cache_seed,
            as_dict=True,
        )
        pattern_lines.append(
            f"    group_manager_args={{\n{llm_config_arg}    }},"
        )

        # Add context variables if present
        if self.agent.data.context_variables:
            # Generate context variables lines
            ctx_lines = self._generate_context_variables()
            if ctx_lines:
                pattern_lines.extend(
                    [
                        "    context_variables=ContextVariables(data={",
                        *ctx_lines,
                        "    }),",
                    ]
                )

        # Add after work if applicable
        if self._should_check_for_after_work(pattern_class):
            after_work_arg, _ = self._get_after_work_configuration()
            if after_work_arg:
                pattern_lines.append(f"    group_after_work={after_work_arg},")

        pattern_lines.append(")")

        return "\n" + "\n".join(pattern_lines)

    def _generate_group_chat_definition(
        self, group_members: list[WaldiezAgent], admin_name: str
    ) -> dict[str, str]:
        """Generate traditional group chat definition."""
        if not isinstance(self.agent, WaldiezGroupManager):
            return {}
        agent_name = self.agent_names[self.agent.id]
        group_chat_name = f"{agent_name}_group_chat"

        group_members_str = ", ".join(
            self.agent_names[member.id] for member in group_members
        )

        lines = [
            f"{group_chat_name} = GroupChat(",
            f"    agents=[{group_members_str}],",
            f"    enable_clear_history={self.agent.data.enable_clear_history},",
            f"    send_introductions={self.agent.data.send_introductions},",
            "    messages=[],",
        ]

        if self.agent.data.max_round > 0:
            lines.append(f"    max_round={self.agent.data.max_round},")

        # Add admin name
        admin = self.agent.data.admin_name or admin_name
        lines.append(f'    admin_name="{admin}",')

        # Add speakers configuration
        speakers_config, custom_selection = (
            self._generate_speakers_configuration()
        )
        lines.extend(speakers_config)

        lines.append(")")

        return {
            "definition": "\n" + "\n".join(lines) + "\n",
            "name": group_chat_name,
            "custom_selection": custom_selection or "",
        }

    def _generate_context_variables(
        self,
    ) -> list[str]:
        """Generate context variables for the group manager agent."""
        ctx_lines: list[str] = []
        for key, value in self.agent.data.context_variables.items():
            if isinstance(value, str):
                ctx_lines.append(f'        "{key}": "{value}",')
            else:
                ctx_lines.append(f'        "{key}": {value},')
        return ctx_lines

    def _generate_speakers_configuration(
        self,
    ) -> tuple[list[str], Optional[str]]:
        """Generate speakers configuration for traditional group chat."""
        if not isinstance(self.agent, WaldiezGroupManager):
            return [], None
        config_lines: list[str] = []
        custom_function = None

        speakers = self.agent.data.speakers

        # Max retries
        if speakers.max_retries_for_selecting is not None:
            config_lines.append(
                "    max_retries_for_selecting_speaker="
                f"{speakers.max_retries_for_selecting},"
            )

        # Selection method
        if speakers.selection_method != "custom":
            config_lines.append(
                f'    speaker_selection_method="{speakers.selection_method}",'
            )
        else:
            # Custom selection method
            agent_name = self.agent_names[self.agent.id]
            custom_function, function_name = (
                speakers.get_custom_method_function(name_suffix=agent_name)
            )
            config_lines.append(
                f"    speaker_selection_method={function_name},"
            )

        # Selection mode configurations
        if speakers.selection_mode == "repeat":
            config_lines.extend(self._generate_repeat_configuration())
        elif speakers.selection_mode == "transition":
            config_lines.extend(self._generate_transition_configuration())

        # Clean up None values
        config_lines = [line.replace('"None"', "None") for line in config_lines]

        return config_lines, custom_function

    def _generate_repeat_configuration(self) -> list[str]:
        """Generate repeat selection configuration."""
        if not isinstance(self.agent, WaldiezGroupManager):
            return []
        lines: list[str] = []
        allow_repeat = self.agent.data.speakers.allow_repeat

        if isinstance(allow_repeat, bool):
            lines.append(f"    allow_repeat_speaker={allow_repeat},")
        else:
            # List of agent names
            agent_names = ", ".join(
                self.agent_names[agent_id] for agent_id in allow_repeat
            )
            lines.append(f"    allow_repeat=[{agent_names}],")

        return lines

    def _generate_transition_configuration(self) -> list[str]:
        """Generate transition selection configuration."""
        if not isinstance(self.agent, WaldiezGroupManager):
            return []
        lines: list[str] = []

        if not self.agent.data.speakers.allowed_or_disallowed_transitions:
            return lines

        # Convert agent IDs to names in transitions
        transitions_dict = {}
        for (
            agent_id,
            transitions,
        ) in self.agent.data.speakers.allowed_or_disallowed_transitions.items():
            agent_name = self.agent_names[agent_id]
            transition_names = [self.agent_names[tid] for tid in transitions]
            transitions_dict[agent_name] = transition_names

        # Serialize transitions
        transitions_str = self.serializer(transitions_dict, 1)
        transitions_str = transitions_str.replace('"', "").replace("'", "")

        lines.extend(
            [
                "    allowed_or_disallowed_speaker_transitions="
                f"{transitions_str},",
                '    speaker_transitions_type="'
                f'"{self.agent.data.speakers.transitions_type}",',
            ]
        )

        return lines

    def _should_check_for_after_work(self, pattern_class: str) -> bool:
        """Check if pattern should have after work configuration."""
        return pattern_class not in ["ManualPattern", "AutoPattern"]

    def _get_after_work_configuration(self) -> tuple[str, str]:
        """Get after work target configuration."""
        if not self.agent.data.after_work:
            return "", ""

        return self._get_transition_target(self.agent.data.after_work)

    def _get_transition_target(
        self, target: WaldiezTransitionTarget
    ) -> tuple[str, str]:
        """Get transition target configuration and import."""
        import_prefix = "from autogen.agentchat.group"

        target_map = {
            "TerminateTarget": (
                "TerminateTarget()",
                f"{import_prefix} import TerminateTarget",
            ),
            "AskUserTarget": (
                "AskUserTarget()",
                f"{import_prefix} import AskUserTarget",
            ),
            "RevertToUserTarget": (
                "RevertToUserTarget()",
                f"{import_prefix} import RevertToUserTarget",
            ),
            "StayTarget": (
                "StayTarget()",
                f"{import_prefix} import StayTarget",
            ),
            "GroupManagerTarget": (
                "GroupManagerTarget()",
                f"{import_prefix}.targets.group_manager_target "
                "import GroupManagerTarget",
            ),
        }

        if target.target_type in target_map:
            return target_map[target.target_type]

        # Special cases
        if target.target_type == "AgentTarget":
            target_name = self.agent_names[target.value[0]]
            return (
                f"AgentTarget(agent={target_name})",
                f"{import_prefix} import AgentTarget",
            )

        if target.target_type == "RandomAgentTarget":
            target_names = [
                self.agent_names[agent_id] for agent_id in target.value
            ]
            if not target_names:
                target_names = [
                    self.agent_names[agent.id]
                    for agent in self.group_chat_members
                ]
            target_names_str = ", ".join(target_names)
            return (
                f"RandomAgentTarget(agents=[{target_names_str}])",
                f"{import_prefix}.targets.transition_target "
                "import RandomAgentTarget",
            )

        return "", ""



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Content merger for combining multiple export results."""

from dataclasses import dataclass, field
from typing import Optional, Union

from ..core import (
    ContentOrder,
    EnvironmentVariable,
    ExporterContext,
    ExportResult,
    ImportStatement,
    PositionedContent,
    ValidationError,
    ValidationResult,
)


@dataclass
class MergeStatistics:
    """Statistics about the merge operation."""

    total_results: int = 0
    total_imports: int = 0
    deduplicated_imports: int = 0
    total_content_items: int = 0
    total_env_vars: int = 0
    deduplicated_env_vars: int = 0
    conflicts_found: list[str] = field(default_factory=list[str])


class ContentMerger:
    """Intelligently merges multiple ExportResult objects."""

    def __init__(self, context: Optional[ExporterContext] = None):
        """Initialize the content merger.

        Parameters
        ----------
        context : Optional[ExporterContext], optional
            The exporter context, by default None
        """
        self.context = context or ExporterContext()
        self.statistics = MergeStatistics()

    def merge_results(self, results: list[ExportResult]) -> ExportResult:
        """Merge multiple export results into one.

        Parameters
        ----------
        results : list[ExportResult]
            The export results to merge

        Returns
        -------
        ExportResult
            The merged export result
        """
        if not results:
            return ExportResult()

        if len(results) == 1:
            return results[0]

        # Initialize statistics
        self.statistics = MergeStatistics(total_results=len(results))

        # Create merged result
        merged = ExportResult()

        # 1. Merge imports with intelligent deduplication
        merged.imports = self._merge_imports(results)

        # 2. Merge positioned content with proper ordering
        merged.positioned_content = self._merge_positioned_content(results)

        # 3. Merge environment variables with conflict detection
        merged.environment_variables = self._merge_environment_variables(
            results
        )

        # 4. Merge validation results
        merged.validation_result = self._merge_validation_results(results)

        # 5. Handle main content (typically empty for merged results)
        merged.main_content = self._merge_main_content(results)

        return merged

    def _merge_imports(
        self, results: list[ExportResult]
    ) -> set[ImportStatement]:
        """Merge imports with intelligent deduplication and prioritization.

        Parameters
        ----------
        results : list[ExportResult]
            The export results containing imports

        Returns
        -------
        set[ImportStatement]
            The merged and deduplicated imports
        """
        all_imports: set[ImportStatement] = set()

        # Collect all imports
        for result in results:
            all_imports.update(result.imports)
            self.statistics.total_imports += len(result.imports)

        # Deduplicate based on statement content, keeping best position
        deduplicated = self._deduplicate_imports(all_imports)

        self.statistics.deduplicated_imports = len(deduplicated)

        return deduplicated

    def _deduplicate_imports(
        self, imports: set[ImportStatement]
    ) -> set[ImportStatement]:
        """Deduplicate imports intelligently.

        Rules:
        1. Keep import with highest priority (BUILTINS > THIRD_PARTY > LOCAL)
        2. For same position, keep the one with more metadata
        3. Detect conflicts and warn

        Parameters
        ----------
        imports : set[ImportStatement]
            The imports to deduplicate

        Returns
        -------
        set[ImportStatement]
            The deduplicated imports
        """
        # Group by statement content
        grouped: dict[str, ImportStatement] = {}
        conflicts: list[str] = []

        for imp in imports:
            key = imp.statement.strip()

            if key not in grouped:
                grouped[key] = imp
            else:
                existing = grouped[key]

                # Prioritize by position (lower enum value = higher priority)
                if imp.position.value < existing.position.value:
                    # New import has higher priority position
                    grouped[key] = imp
                elif imp.position.value == existing.position.value:
                    # Same position - check for conflicts
                    if imp.metadata != existing.metadata:
                        conflicts.append(
                            f"Import '{key}' has conflicting metadata: "
                            f"{existing.metadata} vs {imp.metadata}"
                        )
                    # Keep existing (first wins for same priority)

        if conflicts:
            self.statistics.conflicts_found.extend(conflicts)
            # Log conflicts but continue
            for conflict in conflicts:
                self.context.get_logger().warning(
                    f"Import conflict: {conflict}"
                )

        return set(grouped.values())

    def _merge_positioned_content(
        self, results: list[ExportResult]
    ) -> list[PositionedContent]:
        """Merge positioned content with proper ordering.

        The challenge: Different exporters add content at different positions,
        and we need to maintain the correct order while handling a
        gent-specific positioning.

        Parameters
        ----------
        results : list[ExportResult]
            The export results containing positioned content

        Returns
        -------
        list[PositionedContent]
            The merged and properly ordered content
        """
        all_content: list[PositionedContent] = []

        # Collect all positioned content
        for result in results:
            all_content.extend(result.positioned_content)
            self.statistics.total_content_items += len(
                result.positioned_content
            )

        # Sort by complex criteria for proper ordering
        sorted_content = self._sort_positioned_content(all_content)

        return sorted_content

    def _sort_positioned_content(
        self, content: list[PositionedContent]
    ) -> list[PositionedContent]:
        """Sort positioned content by multiple criteria.

        Sorting Priority:
        1. ExportPosition (TOP, IMPORTS, TOOLS, MODELS, AGENTS, CHATS, BOTTOM)
        2. ContentOrder within position (EARLY_SETUP, SETUP, MAIN_CONTENT, etc.)
        3. Agent ID (for agent-specific content)
        4. AgentPosition (BEFORE_ALL, BEFORE, AS_ARGUMENT, AFTER, AFTER_ALL)
        5. Original order (for stability)

        Parameters
        ----------
        content : list[PositionedContent]
            The content to sort

        Returns
        -------
        list[PositionedContent]
            The sorted content
        """
        return sorted(
            content,
            key=lambda pc: (
                pc.position.value,  # 1. ExportPosition enum value
                self._get_order_value(
                    pc.order
                ),  # 2. ContentOrder numeric value
                pc.agent_id
                or "",  # 3. Agent ID (empty string for non-agent content)
                (
                    pc.agent_position.value if pc.agent_position else 0
                ),  # 4. AgentPosition
                id(pc),  # 5. Original object ID for stability
            ),
        )

    # pylint: disable=no-self-use
    def _get_order_value(self, order: Union[int, ContentOrder]) -> int:
        """Get numeric value from order for sorting.

        Parameters
        ----------
        order : Union[int, ContentOrder]
            The order value

        Returns
        -------
        int
            The numeric order value
        """
        return order.value if isinstance(order, ContentOrder) else order

    def _merge_environment_variables(
        self, results: list[ExportResult]
    ) -> list[EnvironmentVariable]:
        """Merge environment variables with conflict detection.

        Rules:
        1. Deduplicate by variable name
        2. Detect value conflicts and warn
        3. Keep first occurrence for conflicts
        4. Preserve metadata and descriptions

        Parameters
        ----------
        results : list[ExportResult]
            The export results containing environment variables

        Returns
        -------
        list[EnvironmentVariable]
            The merged environment variables
        """
        env_vars: dict[str, EnvironmentVariable] = {}
        conflicts: list[str] = []
        for result in results:
            for env_var in result.environment_variables:
                self.statistics.total_env_vars += 1
                key = env_var.name
                if key in env_vars:
                    existing = env_vars[key]

                    # Check for value conflicts
                    if existing.value != env_var.value:
                        conflicts.append(
                            f"Environment variable '{key}' has conflicting "
                            f"values: '{existing.value}' vs '{env_var.value}'"
                        )
                    # Keep first occurrence
                    # (tools/models take precedence over agents)

                else:
                    env_vars[key] = env_var

        self.statistics.deduplicated_env_vars = len(env_vars)

        if conflicts:
            self.statistics.conflicts_found.extend(conflicts)
            # Log conflicts but continue (non-fatal)
            for conflict in conflicts:
                self.context.get_logger().warning(
                    f"Environment variable conflict: {conflict}"
                )

        return list(env_vars.values())

    def _merge_validation_results(
        self, results: list[ExportResult]
    ) -> ValidationResult:
        """Merge validation results from all export results.

        Parameters
        ----------
        results : list[ExportResult]
            The export results containing validation results

        Returns
        -------
        ValidationResult
            The merged validation result
        """
        all_errors: list[ValidationError] = []
        all_warnings: list[ValidationError] = []

        for result in results:
            if result.validation_result:
                all_errors.extend(result.validation_result.errors)
                all_warnings.extend(result.validation_result.warnings)

        # Add merger-specific warnings
        if self.statistics.conflicts_found:
            all_warnings.extend(
                [
                    ValidationError(message=f"Merge conflict: {conflict}")
                    for conflict in self.statistics.conflicts_found
                ]
            )

        return ValidationResult(
            is_valid=not all_errors,
            errors=all_errors,
            warnings=all_warnings,
        )

    def _merge_main_content(self, results: list[ExportResult]) -> Optional[str]:
        """Merge main content from results.

        For flow exports, main content is typically empty since all content
        is positioned. But we handle it for completeness.

        Parameters
        ----------
        results : list[ExportResult]
            The export results

        Returns
        -------
        Optional[str]
            The merged main content, or None
        """
        main_contents: list[str] = []

        for result in results:
            if result.main_content and result.main_content.strip():
                main_contents.append(result.main_content.strip())

        return "\n\n".join(main_contents) if main_contents else None

    def get_merge_statistics(self) -> MergeStatistics:
        """Get statistics about the last merge operation.

        Returns
        -------
        MergeStatistics
            Statistics about the merge operation
        """
        return self.statistics



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Ensure unique names for agents, models, tools, and chats."""

from typing import Iterable, TypedDict

from ..agents import WaldiezAgent
from ..chat import WaldiezChat
from ..common import MAX_VARIABLE_LENGTH, get_valid_instance_name
from ..model import WaldiezModel
from ..tool import WaldiezTool


class WaldiezUniqueNames(TypedDict):
    """The result type for ensure_unique_names."""

    agent_names: dict[str, str]
    model_names: dict[str, str]
    tool_names: dict[str, str]
    chat_names: dict[str, str]
    agents: list[WaldiezAgent]
    models: list[WaldiezModel]
    tools: list[WaldiezTool]
    chats: list[WaldiezChat]
    flow_name: str


# pylint: disable=too-many-locals
def ensure_unique_names(
    flow_id: str,
    flow_name: str,
    flow_agents: Iterable[WaldiezAgent],
    flow_models: Iterable[WaldiezModel],
    flow_tools: Iterable[WaldiezTool],
    flow_chats: Iterable[WaldiezChat],
    max_length: int = MAX_VARIABLE_LENGTH,
    flow_name_max_length: int = 20,
) -> WaldiezUniqueNames:
    """Ensure unique names for agents, models, tools, and chats and flow.

    Parameters
    ----------
    flow_id : str
        The ID of the flow.
    flow_name : str
        The name of the flow.
    flow_agents : Iterable[WaldiezAgent]
        The agents in the flow.
    flow_models : Iterable[WaldiezModel]
        The models in the flow.
    flow_tools : Iterable[WaldiezTool]
        The tools in the flow.
    flow_chats : Iterable[WaldiezChat]
        The chats in the flow.
    max_length : int, optional
        The maximum length of the name, by default 46
    flow_name_max_length : int, optional
        The maximum length of the flow name, by default 20

    Returns
    -------
    ResultType
        The result with unique names for agents, models, tools, chats, flow.
    """
    all_names: dict[str, str] = {}
    agent_names: dict[str, str] = {}
    model_names: dict[str, str] = {}
    tool_names: dict[str, str] = {}
    chat_names: dict[str, str] = {}
    agents: list[WaldiezAgent] = []
    models: list[WaldiezModel] = []
    tools: list[WaldiezTool] = []
    chats: list[WaldiezChat] = []

    for agent in flow_agents:
        all_names = get_valid_instance_name(
            (agent.id, agent.name),
            all_names,
            prefix="wa",
            max_length=max_length,
        )
        agent_names[agent.id] = all_names[agent.id]
        agents.append(agent)

    for model in flow_models:
        all_names = get_valid_instance_name(
            (model.id, model.name),
            all_names,
            prefix="wm",
            max_length=max_length,
        )
        model_names[model.id] = all_names[model.id]
        models.append(model)

    for tool in flow_tools:
        all_names = get_valid_instance_name(
            (tool.id, tool.name),
            all_names,
            prefix="ws",
            max_length=max_length,
        )
        tool_names[tool.id] = all_names[tool.id]
        tools.append(tool)

    for chat in flow_chats:
        all_names = get_valid_instance_name(
            (chat.id, chat.name), all_names, prefix="wc", max_length=max_length
        )
        chat_names[chat.id] = all_names[chat.id]
        chats.append(chat)

    all_names = get_valid_instance_name(
        (flow_id, flow_name),
        all_names,
        prefix="wf",
        max_length=flow_name_max_length,
    )
    flow_name = all_names[flow_id]

    result: WaldiezUniqueNames = {
        "agent_names": agent_names,
        "model_names": model_names,
        "tool_names": tool_names,
        "chat_names": chat_names,
        "agents": agents,
        "models": models,
        "tools": tools,
        "chats": chats,
        "flow_name": flow_name,
    }
    return result



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=unused-argument
"""Text media content models."""

from pathlib import Path

from pydantic import BaseModel
from typing_extensions import Literal


class TextMediaContent(BaseModel):
    """Text media content."""

    type: Literal["text"] = "text"
    text: str

    def to_string(
        self,
        uploads_root: Path | None = None,
        base_name: str | None = None,
    ) -> str:
        """Convert the content to a string.

        Parameters
        ----------
        uploads_root : Path | None
            The root directory for storing images, optional.
        base_name : str | None
            The base name for the image file, optional.

        Returns
        -------
        str
            The string representation of the content.
        """
        return self.text



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=unused-argument
"""File media content models."""

from pathlib import Path
from typing import Literal

from pydantic import BaseModel

from .base import FileContent


class FileMediaContent(BaseModel):
    """File media content."""

    type: Literal["file"] = "file"
    file: FileContent

    def to_string(
        self,
        uploads_root: Path | None,
        base_name: str | None = None,
    ) -> str:
        """Convert the content to a string.

        Parameters
        ----------
        uploads_root : Path | None
            The root directory for storing images, optional.
        base_name : str | None
            The base name for the image file, optional.

        Returns
        -------
        str
            The string representation of the content.
        """
        if self.file.url:
            return f"<a href='{self.file.url}'>{self.file.name}</a>"
        if self.file.file:
            return f"<a href='{self.file.file}'>{self.file.name}</a>"
        return str(self.file.file)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Flow exporter."""

from pathlib import Path
from typing import Any, Optional

from waldiez.models import Waldiez

from ..core import Exporter, ExporterContext
from ..core.extras import FlowExtras
from .file_generator import FileGenerator
from .orchestrator import ExportOrchestrator


class FlowExporter(Exporter[FlowExtras]):
    """Flow exporter."""

    def __init__(
        self,
        waldiez: Waldiez,
        output_dir: Path | None,
        for_notebook: bool,
        context: Optional[ExporterContext] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the chats exporter.

        Parameters
        ----------
        waldiez : Waldiez
            The Waldiez instance containing the flow data.
        output_dir : Path
            The directory where the exported flow will be saved.
        for_notebook : bool
            Whether the export is intended for a notebook environment.
        context : Optional[ExporterContext], optional
            Exporter context with dependencies, by default None
        **kwargs : Any
            Additional keyword arguments for the exporter.
        """
        super().__init__(context, **kwargs)

        self.waldiez = waldiez
        self.output_dir = Path(output_dir) if output_dir is not None else None
        self.flow_config = self.context.get_config(
            name=waldiez.name,
            description=waldiez.description,
            requirements=waldiez.requirements or [],
            tags=waldiez.tags or [],
            output_extension="ipynb" if for_notebook else "py",
            is_async=waldiez.is_async,
            output_directory=str(self.output_dir) if self.output_dir else None,
            cache_seed=waldiez.cache_seed,
        )
        self._extras = self._generate_extras()

    @property
    def extras(self) -> FlowExtras:
        """Get the flow exporter extras.

        Returns
        -------
        dict[str, Any]
            The extras dictionary containing additional
            information for the flow exporter.
        """
        return self._extras

    def _generate_extras(self) -> FlowExtras:
        """Generate the extras for the flow exporter.

        Returns
        -------
        BaseExtras
            An instance of BaseExtras containing the generated content.
        """
        extras = FlowExtras(
            instance_id=self.waldiez.id,
            flow_name=self.waldiez.name,
            description=self.waldiez.description or "",
            config=self.flow_config,
        )
        return extras

    def generate_main_content(self) -> str:
        """Generate the main content of the export.

        Returns
        -------
        str
            The final executable script or notebook content.
        """
        orchestrator = ExportOrchestrator(
            waldiez=self.waldiez,
            context=self.context,
        )
        merged_result = orchestrator.orchestrate()
        after_run = orchestrator.get_after_run_content()
        generator = FileGenerator(
            context=self.context,
        )
        return generator.generate(
            merged_result=merged_result,
            is_async=self.waldiez.is_async,
            after_run=after_run,
        )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""LLM config argument utility functions for Waldiez agents."""

from waldiez.models import WaldiezAgent, WaldiezModel


def get_agent_llm_config_arg(
    agent: WaldiezAgent,
    all_models: list[WaldiezModel],
    model_names: dict[str, str],
    cache_seed: int | None,
    as_dict: bool = False,
    tabs: int = 1,
) -> str:
    """Get the string representation of the agent's llm config argument.

    Parameters
    ----------
    agent : WaldiezAgent
        The agent.
    all_models : list[WaldiezModel]
        All the models in the flow.
    model_names : dict[str, str]
        A mapping of model ids to model names.
    cache_seed : Optional[int]
        The cache seed.
    as_dict : bool, optional
        Whether to return the argument as a dictionary, by default False.
    tabs : int, optional
        The number of tabs for indentation, by default 1.

    Returns
    -------
    str
        The agent's llm config argument to use.
    """
    if as_dict is False:
        return _get_agent_llm_config_arg_as_arg(
            agent,
            all_models,
            model_names,
            cache_seed,
            tabs=tabs,
        )
    return _get_agent_llm_config_arg_as_dict(
        agent,
        all_models,
        model_names,
        cache_seed,
        tabs=tabs + 1,
    )


def _get_agent_llm_config_arg_as_arg(
    agent: WaldiezAgent,
    all_models: list[WaldiezModel],
    model_names: dict[str, str],
    cache_seed: int | None,
    tabs: int = 1,
    tab_leng: int = 4,
) -> str:
    tab = " " * tab_leng * tabs if tabs > 0 else ""
    # tab = "    " * tabs if tabs > 0 else ""
    if not agent.data.model_ids:
        return f"{tab}llm_config=False,  # pyright: ignore" + "\n"
    content = f"{tab}llm_config=autogen.LLMConfig(" + "\n"
    content += f"{tab}    config_list=["
    got_at_least_one_model = False
    temperature: float | None = None
    for model_id in agent.data.model_ids:
        model = next((m for m in all_models if m.id == model_id), None)
        if model is not None:
            temperature = model.data.temperature
            model_name = model_names[model_id]
            content += "\n" + f"{tab}        {model_name}_llm_config,"
            got_at_least_one_model = True
    if not got_at_least_one_model:  # pragma: no cover
        return f"{tab}llm_config=False,  # pyright: ignore" + "\n"
    content += "\n" + f"{tab}    ]," + "\n"
    content += f"{tab}    cache_seed={cache_seed}," + "\n"
    if temperature is not None:
        content += f"{tab}    temperature={temperature}," + "\n"
    content += tab + "),\n"
    return content


def _get_agent_llm_config_arg_as_dict(
    agent: WaldiezAgent,
    all_models: list[WaldiezModel],
    model_names: dict[str, str],
    cache_seed: int | None,
    tabs: int = 1,
    tab_leng: int = 4,
) -> str:
    tab = " " * tab_leng * tabs if tabs > 0 else ""
    if not agent.data.model_ids:
        return f'{tab}"llm_config": False' + "\n"
    content = f'{tab}"llm_config": autogen.LLMConfig(' + "\n"
    content += f"{tab}    config_list=["
    got_at_least_one_model = False
    temperature: float | None = None
    for model_id in agent.data.model_ids:
        model = next((m for m in all_models if m.id == model_id), None)
        if model is not None:  # pragma: no branch
            temperature = model.data.temperature
            model_name = model_names[model_id]
            content += "\n" + f"{tab}        {model_name}_llm_config,"
            got_at_least_one_model = True
    if not got_at_least_one_model:  # pragma: no cover
        return f'{tab}"llm_config": False' + "\n"
    content += "\n" + f"{tab}    ]," + "\n"
    content += f"{tab}    cache_seed={cache_seed}," + "\n"
    if temperature is not None:
        content += f"{tab}    temperature={temperature}," + "\n"
    content += tab + "),\n"
    return content



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-few-public-methods,no-self-use
"""Captain agent configuration processor."""

import json
import os
import uuid
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Optional

from waldiez.models import (
    WaldiezAgent,
    WaldiezCaptainAgent,
    WaldiezModel,
    WaldiezModelData,
)

from ...core import (
    CodeExecutionConfig,
    DefaultSerializer,
    Serializer,
    SystemMessageConfig,
    TerminationConfig,
)
from ...core.extras.agent_extras import CaptainExtras


class CaptainAgentProcessor:
    """Processor for captain agent configuration."""

    def __init__(
        self,
        agent: WaldiezAgent,
        agent_names: dict[str, str],
        all_models: list[WaldiezModel],
        serializer: Optional[Serializer] = None,
        output_dir: Optional[Path] = None,
    ):
        self.agent = agent
        self.agent_names = agent_names
        self.all_models = all_models
        self.serializer = serializer or DefaultSerializer()
        self.output_dir = output_dir

    def process(
        self,
        code_execution_config: Optional[CodeExecutionConfig] = None,
        termination_config: Optional[TerminationConfig] = None,
        system_message_config: Optional[SystemMessageConfig] = None,
    ) -> CaptainExtras:
        """Process group manager and return extras.

        Parameters
        ----------
        code_execution_config : CodeExecutionConfig, optional
            Configuration for code execution, by default None.
        termination_config : TerminationConfig, optional
            Configuration for termination, by default None.
        system_message_config : SystemMessageConfig, optional
            Configuration for system messages, by default None.

        Returns
        -------
        CaptainExtras
            The processed result containing extra arguments, before content,
            imports, and environment variables.
        """
        result = CaptainExtras(
            instance_id=self.agent.id,
            code_execution_config=code_execution_config,
            termination_config=termination_config,
            system_message_config=system_message_config,
        )
        if not self.agent.is_captain or not isinstance(
            self.agent, WaldiezCaptainAgent
        ):
            return result
        agent_name = self.agent_names[self.agent.id]
        save_path = str(self.output_dir) if self.output_dir else "."
        if save_path != ".":
            os.makedirs(save_path, exist_ok=True)
        result.add_arg("agent_config_save_path=os.getcwd()", tabs=1)
        if self.agent.data.agent_lib:
            lib_dict = [
                lib.model_dump(by_alias=False)
                for lib in self.agent.data.agent_lib
            ]
            lib_json_name = f"{agent_name}_agent_lib.json"
            agent_lib_path = os.path.join(save_path, lib_json_name)
            with open(agent_lib_path, "w", encoding="utf-8", newline="\n") as f:
                json.dump(lib_dict, f, ensure_ascii=False, indent=4)
            result.add_arg(f'agent_lib="{lib_json_name}"', tabs=1)
        if self.agent.data.tool_lib:
            result.add_arg(f'tool_lib="{self.agent.data.tool_lib}"', tabs=1)
        nested_config = self._generate_nested_config(
            self.agent,
            agent_name,
            save_path,
        )
        result.set_nested_config(nested_config)
        serialized_nested_config = self.serializer.serialize(nested_config)
        result.add_arg(f"nested_config={serialized_nested_config}", tabs=1)
        return result

    def _generate_nested_config(
        self,
        agent: WaldiezCaptainAgent,
        agent_name: str,
        save_path: str,
    ) -> dict[str, Any]:
        """Generate the nested config for the captain agent.

        Parameters
        ----------
        agent_name : str
            The agent name.
        save_path : str
            The path to save the nested config.

        Returns
        -------
        dict[str, Any]
            The nested config.
        """
        config_file_or_env_name = f"{agent_name}_llm_config.json"
        llm_config_list = self._get_llm_configs()
        os.makedirs(save_path, exist_ok=True)
        config_file_or_env_path = os.path.join(
            save_path, config_file_or_env_name
        )
        with open(
            config_file_or_env_path, "w", encoding="utf-8", newline="\n"
        ) as f:
            json.dump(llm_config_list, f, ensure_ascii=False, indent=4)
        llm_config = llm_config_list[0]
        if "temperature" not in llm_config:
            llm_config["temperature"] = 1
        if "top_p" not in llm_config:
            llm_config["top_p"] = 0.95
        if "max_tokens" not in llm_config:
            llm_config["max_tokens"] = 2048
        return {
            "autobuild_init_config": {
                "config_file_or_env": config_file_or_env_name,
                "builder_model": llm_config["model"],
                "agent_model": llm_config["model"],
            },
            "autobuild_build_config": self._get_auto_build_build_config(
                llm_config
            ),
            "group_chat_config": {"max_round": agent.data.max_round},
            "group_chat_llm_config": None,
            "max_turns": agent.data.max_turns,
        }

    def _get_llm_configs(self) -> list[dict[str, Any]]:
        """Get the LLM configurations for the captain agent.

        Returns
        -------
        list[dict[str, Any]]
            The list of LLM configurations.
        """
        temperature = 1
        top_p = 0.95
        max_tokens = 2048
        models_in_list: list[WaldiezModel] = []
        config_list: list[dict[str, Any]] = []
        for model_id in self.agent.data.model_ids:
            model = self._get_waldiez_model(model_id)
            if model not in models_in_list:
                models_in_list.append(model)
                llm_config = model.get_llm_config(skip_price=True)
                config_list.append(llm_config)
        if not config_list:
            default_model = self._get_default_model(uuid.uuid4().hex)
            default_llm_config = default_model.get_llm_config(skip_price=True)
            if "temperature" not in default_llm_config:
                default_llm_config["temperature"] = temperature
            if "top_p" not in default_llm_config:
                default_llm_config["top_p"] = top_p
            if "max_tokens" not in default_llm_config:
                default_llm_config["max_tokens"] = max_tokens
            config_list.append(default_llm_config)
        return config_list

    def _get_auto_build_build_config(
        self,
        llm_config: dict[str, Any],
    ) -> dict[str, Any]:
        """Get the auto build build config.

        Parameters
        ----------
        llm_config : dict[str, Any]
            The LLM configuration.

        Returns
        -------
        dict[str, Any]
            The auto build build config.
        """
        coding = False
        code_execution_config: dict[str, Any] = {
            "timeout": 300,
            "work_dir": "groupchat",
            "last_n_messages": 1,
            "use_docker": False,
        }
        if self.agent.data.code_execution_config is not False:
            coding = True
            code_execution_config["work_dir"] = (
                self.agent.data.code_execution_config.work_dir or "groupchat"
            )
            code_execution_config["last_n_messages"] = (
                self.agent.data.code_execution_config.last_n_messages or 1
            )
            code_execution_config["timeout"] = (
                self.agent.data.code_execution_config.timeout or 300
            )
        return {
            "default_llm_config": {
                "temperature": llm_config["temperature"],
                "top_p": llm_config["top_p"],
                "max_tokens": llm_config["max_tokens"],
            },
            "code_execution_config": code_execution_config,
            "coding": coding,
        }

    def _get_waldiez_model(self, model_id: str) -> WaldiezModel:
        """Get the Waldiez model by its ID.

        Parameters
        ----------
        model_id : str
            The model's ID.

        Returns
        -------
        WaldiezModel
            The Waldiez model.
        """
        for model in self.all_models:
            if model.id == model_id:
                return model
        return self._get_default_model(model_id)

    def _get_default_model(self, model_id: str) -> WaldiezModel:
        """Get the default model.

        Parameters
        ----------
        model_id : str
            The model's id.

        Returns
        -------
        WaldiezModel
            The default model.
        """
        now = (
            datetime.now(tz=timezone.utc)
            .isoformat(timespec="milliseconds")
            .replace("+00:00", "Z")
        )
        return WaldiezModel(
            id=model_id,
            type="model",
            name="gpt-4o",
            description="The GPT-4o model.",
            tags=["gpt-4o"],
            requirements=[],
            created_at=now,
            updated_at=now,
            data=WaldiezModelData(
                api_type="openai",
                temperature=1,
                top_p=0.95,
                max_tokens=2048,
                base_url=None,
                api_key=None,
                api_version=None,
                default_headers={},
                price=None,
            ),
        )


# DEFAULT_NESTED_CONFIG = {
#         "autobuild_init_config": {
#             "config_file_or_env": "OAI_CONFIG_LIST",
#             "builder_model": "gpt-4o",
#             "agent_model": "gpt-4o",
#         },
#         "autobuild_build_config": {
#             "default_llm_config": {
#                   "temperature": 1,
#                   "top_p": 0.95,
#                   "max_tokens": 2048
#             },
#             "code_execution_config": {
#                 "timeout": 300,
#                 "work_dir": "groupchat",
#                 "last_n_messages": 1,
#                 "use_docker": False,
#             },
#             "coding": True,
#         },
#         "group_chat_config": {"max_round": 10},
#         "group_chat_llm_config": None,
#         "max_turns": 5,
#     }



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Read the version from the version file."""

from pathlib import Path


def get_waldiez_version() -> (
    str
):  # pragma: no cover  # depends on file existence
    """Read the version from the version file.

    Returns
    -------
    str
        The version.
    """
    here = Path(__file__).parent  # waldiez/models/common
    package_dir = here.parent.parent  # waldiez/models -> waldiez
    version_file = package_dir / "_version.py"
    if not version_file.exists():
        return "0.0.0"  # dev / ignored
    with version_file.open() as f:
        for line in f:
            if line.startswith("__version__"):
                version = line.split("=")[1].strip().strip('"').strip("'")
                # send version without "v" prefix
                if version.startswith("v"):
                    version = version[1:]
                # make sure it is a valid semver
                if not version or not all(
                    part.isdigit() for part in version.split(".")
                ):
                    return "0.0.0"
                return version
    return "0.0.0"  # fallback if not found



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Agents exporter extras."""

from .captain_agent_extras import CaptainAgentProcessor
from .group_manager_agent_extas import GroupManagerProcessor
from .group_member_extras import GroupMemberAgentProcessor
from .reasoning_agent_extras import ReasoningAgentProcessor

__all__ = [
    "CaptainAgentProcessor",
    "GroupManagerProcessor",
    "GroupMemberAgentProcessor",
    "ReasoningAgentProcessor",
]
